# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import Protocol
from typing import TypeVar

import pydantic


T = TypeVar('T', bound=pydantic.BaseModel)


class IPolymorphicCursor(Protocol, Generic[T]):
    items: list[T] = []
    token: str | None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections

from .models import Subject
from .types import ISubjectRepository


class MemorySubjectRepository(ISubjectRepository):
    __module__: str = 'cbra.core.iam'
    objects: dict[str, dict[str, Subject]] = collections.defaultdict(dict)

    @classmethod
    def clear(cls) -> None:
        cls.objects = collections.defaultdict(dict)

    def __init__(self):
        self.objects = MemorySubjectRepository.objects


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
from google.cloud.datastore import Client
from ckms.core import Keychain

from cbra.ext.security import ApplicationKeychain
from cbra.ext.google.environ import GOOGLE_DATASTORE_NAMESPACE
from cbra.ext.google.environ import GOOGLE_SERVICE_PROJECT


@pytest.fixture(scope='session')
def client() -> Client:
    if not GOOGLE_SERVICE_PROJECT or not GOOGLE_DATASTORE_NAMESPACE:
        pytest.skip("Google not configured")
    return Client(
        project=GOOGLE_SERVICE_PROJECT,
        namespace=GOOGLE_DATASTORE_NAMESPACE
    )


@pytest.fixture(scope='session')
def keychain():
    return ApplicationKeychain(Keychain())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from cbra.types import IAuthorizationContext
from cbra.types import PolicyPrincipal
from .condition import Condition


class Binding(pydantic.BaseModel):
    role: str = pydantic.Field(
        default=...,
        title="Role",
        description=(
            "Role that is assigned to the list of `members`, or principals. "
            "For example, `roles/viewer`, `roles/editor`, or `roles/owner`."
        )
    )

    members: list[PolicyPrincipal] = pydantic.Field(
        default=[],
        title="Members",
        description=(
            "Specifies the principals requesting access for a resource. "
            "The `members` array can have the following values:\n\n"
            "- **user:{email}**: an email address that is associated to "
            "a specific account."
        )
    )

    condition: Condition | None = pydantic.Field(
        default=None,
        title="Condition",
        description=(
            "The condition that is associated with this binding.\n\n"
            "If the condition evaluates to `true`, then this binding "
            "applies to the current request.\n\n"
            "If the condition evaluates to `false`, then this binding "
            "does not apply to the current request. However, a different "
            "role binding might grant the same role to one or more "
            "of the principals in this binding.\n\n"
        )
    )

    def has(self, member: PolicyPrincipal):
        """Return a boolean indicating if this binding has the given member."""
        return member in self.members

    def is_conditional(self) -> bool:
        """Return a boolean indicating if the binding is conditional."""
        return self.condition is not None

    def is_satisfied(self, ctx: IAuthorizationContext) -> bool:
        """Return a boolean if the binding is satisfied by the authorization
        context.
        """
        if self.condition:
            raise NotImplementedError
        return any([self.has(p) for p in ctx.get_principals()])

    def update(self, members: set[PolicyPrincipal]):
        """Updates the :class:`Binding` with the given members."""
        for member in members:
            if self.has(member):
                continue
            self.members.append(member)



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import TypeVar
from inspect import Signature


T = TypeVar('T', bound='MutableSignature')


class MutableSignature:
    """Exposes an interface to mutate signatures."""
    __module__: str = 'cbra.types'

    @classmethod
    def fromfunction(cls: type[T], func: Any) -> T:
        return cls(func, inspect.signature(func))

    @property
    def named(self) -> list[inspect.Parameter]:
        """The list of named arguments that are not ``self``
        or variable.
        """
        params = list(self.sig.parameters.values())
        return [x for x in params if x.name not in {'args', 'kwargs', 'self'}]

    @property
    def return_annotation(self) -> Any:
        return self.sig.return_annotation

    def __init__(self, obj: Any, signature: Signature):
        self.obj = obj
        self.sig = signature

    def clear(self) -> None:
        """Remote the positional and keyword arguments from the
        signature.
        """
        pass

    def has_param(self, name: str) -> bool:
        """Return a boolean indicating if the signature has the given
        named parameter.
        """
        return name in self.sig.parameters


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from cbra.types import PolicyPrincipal


class Member(pydantic.BaseModel):
    role: str = pydantic.Field(
        default=...,
        title="Role",
        description=(
            "Role that is assigned to `member`, or principal. "
            "For example, `roles/viewer`, `roles/editor`, or `roles/owner`."
        )
    )

    member: PolicyPrincipal = pydantic.Field(
        default=...,
        title="Member",
        description="The principal to which the `role` is granted."
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import operator
import re
from collections import defaultdict

from .basepermissionsfinder import BasePermissionsFinder


class StaticPermissionsFinder(BasePermissionsFinder):
    __module__: str = 'cbra.ext.rbac'
    available: set[str] = set()
    roles: dict[str, set[str]] = defaultdict(set)

    def __init__(self, available: set[str], roles: dict[str, set[str]]):
        self.available = available
        self.roles = roles

    async def get(self, roles: set[str]) -> set[str]:
        """Return the set of permissions granted to the given roles."""
        permissions: set[str] = functools.reduce(
            operator.or_,
            [self.roles.get(x) or set() for x in roles],
            set()
        )
        for permission in list(permissions):
            if '*' not in permission:
                continue
            pattern = re.compile(re.escape(permission).replace('\\*', '[a-z]+'))
            permissions.remove(permission)
            permissions.update({x for x in self.available if pattern.match(x)})
        return permissions



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
from typing import Any
from typing import TypeVar

from .dependency import Dependency
from .dependencynotsatisfied import DependencyNotSatisfied


T = TypeVar('T', bound='Container')


class Container:
    """Holds injected dependencies."""
    __module__: str = 'cbra.core.ioc'
    injected: dict[str, Any] = collections.OrderedDict()
    requirements: set[str]

    @classmethod
    def fromsettings(cls: type[T]) -> T:
        from cbra.core.conf import settings
        container = cls()
        for dependency in getattr(settings, 'DEPENDENCIES', []):
            container.add(Dependency.parse_obj(dependency))
        return container

    def __init__(self):
        self.injected = collections.OrderedDict()
        self.requirements = set()

    def add(self, dependency: Dependency):
        dependency.resolve()
        self.injected[dependency.name] = dependency

    def has(self, name: str) -> bool:
        return name in self.injected

    def inject(self, name: str, value: Any) -> None:
        """Inject a dependency into the container."""
        self.injected[name] = Dependency(
            name=name,
            qualname='',
            symbol=value,
            singleton=True
        )

    def provide(self, name: str, spec: dict[str, Any]) -> None:
        self.injected[name] = Dependency.parse_obj({**spec, 'name': name})
        self.injected[name].resolve()

    def require(self, name: str) -> Dependency:
        if name not in self.injected:
            raise DependencyNotSatisfied(name)
        return self.injected[name]


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets

import cbra.core as cbra
from .book import Book
from .bookpublication import BookPublication


class BookResource(
    cbra.Resource,
    cbra.Create,
    cbra.Delete,
    cbra.Replace,
    cbra.Retrieve,
    cbra.Update,
    model=Book
):
    books: dict[int, Book] = {
        1: Book(
            id=1,
            title="The Hitchhiker's Guide to the Galaxy",
            publications=[
                BookPublication(country_code='UK', published=datetime.date(1979, 10, 12)),
                BookPublication(country_code='US', published=datetime.date(1980, 10, 1)),
            ]
        )
    }

    async def can_create(self, resource: Book) -> bool:
        return not any([x.title == resource.title for x in self.books.values()])

    async def delete(self, resource: Book):
        assert resource.id is not None
        self.books.pop(resource.id)

    async def get_object(self) -> Book | None:
        return self.books.get(int(self.request.path_params['book_id']))

    async def persist(self, resource: Book, create: bool = False) -> Book:
        params = resource.dict()
        if create:
            params['id'] = secrets.choice(range(1000, 9999))
        resource = self.books[params['id']] = Book.parse_obj(params)
        return resource


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.core.conf import settings
from ..types import OIDCProvider


class ClientProvider:
    clients: dict[str, OIDCProvider] = {}

    def __init__(self):
        for client in settings.OAUTH2_CLIENTS:
            provider = OIDCProvider.parse_obj({**client, 'protocol': 'oidc'})
            self.clients[provider.name] = provider

    async def get(self, client_id: str) -> OIDCProvider | None:
        return self.clients.get(client_id)


LocalClientProvider: ClientProvider = fastapi.Depends(ClientProvider)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from cbra.types import Abortable


class InvalidGrant(Abortable):
    __module__: str = 'cbra.ext.oauth2.types'
    error: str = 'invalid_grant'
    message: str = (
        "The provided authorization grant (e.g., authorization code, "
        "resource owner credentials) or refresh token is invalid, expired, "
        "revoked, does not match the redirect URI used in the authorization "
        "request, or was issued to another client."
    )
    status_code: int = 400

    def __init__(self, message: str | None = None):
        self.message = message or self.message

    async def as_response(self) -> fastapi.Response:
        return fastapi.responses.JSONResponse(
            status_code=self.status_code,
            content={
                'error': self.error,
                'error_description': self.message
            }
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeAlias
from typing import Union

from .orderevent import OrderEvent
from .picklistevent import PicklistEvent
from .picklistshipmentevent import PicklistShipmentEvent
from .productevent import ProductEvent
from .purchaseorderevent import PurchaseOrderEvent


__all__: list[str] = [
    'OrderEvent',
    'PicklistEvent',
    'PicklistShipmentEvent',
    'PicqerWebhookMessage',
    'ProductEvent',
    'PurchaseOrderEvent',
]


PicqerWebhookMessage: TypeAlias = Union[
    OrderEvent,
    PicklistEvent,
    PicklistShipmentEvent,
    ProductEvent,
    PurchaseOrderEvent
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Generator
from typing import TypeVar

import fastapi
import fastapi.params

T = TypeVar('T', bound='IDependant')


class IDependant:
    __module__: str = 'cbra.types'

    @classmethod
    def depends(cls: type[T]) -> T:
        """Return a :class:`fastapi.params.Depends` object."""
        return fastapi.Depends(cls.__inject__())

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        """Return a callable that specifies the dependencies of this
        class in its signature.
        """
        return cls
    
    def dependency_resolvers(
        self
    ) -> Generator[Callable[..., Any], None, None]:
        for attname, member in inspect.getmembers(self):
            if not isinstance(member, fastapi.params.Depends):
                continue
            async def f(dep: Any = member):
                setattr(self, attname, dep)
            yield f


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone 
from typing import Any
from typing import AsyncIterable

import pytest
import pytest_asyncio
from ckms.core import Keychain
from headless.core import httpx
from headless.ext.oauth2 import Client

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.core.iam.models import Subject
from cbra.types import Session
from cbra.ext import oauth2


SUPPORTED_RESPONSE_TYPES: list[oauth2.types.ResponseType] = [
    oauth2.types.ResponseType.code
]


class ServerClient(httpx.Client):

    async def login(self, sub: int = 1) -> None:
        key = await cbra.SecretKey.setup()
        session = Session.new()
        session.update({
            'iss': 'https://cbra',
            'sub': sub
        })
        await session.sign(key.sign)
        self.cookies[settings.SESSION_COOKIE_NAME] = session.as_cookie()


@pytest.fixture # type: ignore
def app(issuer: str) -> cbra.Application:
    app = cbra.Application()
    app.add(
        oauth2.AuthorizationServer(
            iss=issuer,
            response_types=[
                oauth2.types.ResponseType.code
            ]
        )
    )
    return app


@pytest.fixture(scope='function')
def storage(
    app: cbra.Application
):
    impl = app.container.require('AuthorizationServerStorage')
    impl.resolve()
    yield impl.get()()


@pytest.fixture
def application_client() -> oauth2.models.Client:
    return oauth2.models.Client.parse_obj({
        'kind': 'Application',
        'client_id': 'app',
        'info': {
            'organization_name': "ACME",
            'display_name': "CBRA Test Client",
        },
        'allowed_redirect_uris': ["https://cbra/oauth2/callback"]
    })


@pytest_asyncio.fixture # type: ignore
async def idp(
    app: cbra.Application
) -> AsyncIterable[Client]:
    params: dict[str, Any] = {
        'app': app,
        'issuer': 'https://cbra'
    }
    async with Client(**params) as client:
        yield client


@pytest_asyncio.fixture # type: ignore
async def client(
    app: cbra.Application,
    storage: oauth2.types.IAuthorizationServerStorage,
    application_client: oauth2.models.Client,
    issuer: str
) -> AsyncIterable[ServerClient]:
    params: dict[str, Any] = {
        'app': app,
        'base_url': issuer
    }
    await storage.persist(application_client)
    async with ServerClient(**params) as client:
        yield client


@pytest_asyncio.fixture # type: ignore
async def evil_client(
    app: cbra.Application,
    storage: oauth2.types.IAuthorizationServerStorage,
    application_client: oauth2.models.Client,
    subject: Subject
) -> AsyncIterable[httpx.Client]:
    class EvilSecretKey(cbra.SecretKey):
        keychain: Keychain = Keychain()
    key = await EvilSecretKey.setup('evil')
    session = Session.new()
    await session.sign(key.sign)
    params: dict[str, Any] = {
        'app': app,
        'base_url': 'https://cbra',
        'cookies': {
            settings.SESSION_COOKIE_NAME: session.as_cookie()
        }
    }
    await storage.persist(application_client)
    await storage.persist(subject)
    async with httpx.Client(**params) as client:
        yield client


@pytest.fixture
def authorization_endpoint() -> str:
    return '/authorize'


@pytest.fixture
def issuer() -> str:
    return 'https://cbra'


@pytest.fixture
def subject() -> Subject:
    now = datetime.now(timezone.utc)
    return Subject(
        kind='User',
        uid=1,
        created=now,
        seen=now,
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import Generic
from typing import TypeVar

from google.cloud.datastore import Client

import cbra.core as cbra
from cbra.ext.security import ApplicationKeychain
from cbra.types import PersistedModel
from .basemodelrepository import BaseModelRepository


T = TypeVar('T', bound=PersistedModel)


class TypedDatastoreRepository(BaseModelRepository, Generic[T]):
    __module__: str = 'cbra.ext.google'
    keychain: ApplicationKeychain | None = None
    model: type[T]

    def __init__(
        self,
        client: Client | Any = cbra.inject('GoogleDatastoreClient'),
        keychain: ApplicationKeychain | None = ApplicationKeychain.depends()
    ):
        super().__init__(client=client)
        self.keychain = keychain

    async def allocate_pk(self) -> int:
        return await super().allocate(self.get_entity_name(self.model))

    async def one(
        self,
        filters: list[tuple[str, str, Any]],
        ordering: list[str] | None = None
    ) -> T | None:
        """Find a single entity using the search predicate."""
        q = self.query(kind=self.get_entity_name(self.model))
        if ordering:
            q.order = ordering
        for attname, op, value in filters:
            q.add_filter(attname, op, value)
        obj: T | None = None
        execute = functools.partial(q.fetch, limit=2)
        for entity in (await self.run_in_executor(execute)):
            if obj is not None:
                raise NotImplementedError("Multiple objects returned")
            obj = self.restore(self.model, entity)
            break
        return obj

    async def get(self, pk: int | str) -> None | T:
        """Retrieve an instance of model by its primary key, or return
        ``None`` if it does not exist.
        """
        e = await self.get_entity_by_key(self.model_key(self.model, pk=pk))
        instance = None
        if e is not None:
            instance = self.restore(self.model, e)
        return instance

    async def persist(self, obj: T, parent: PersistedModel | None = None) -> T:
        """Persist an instance of model."""
        return await self.persist_model(obj, parent=parent)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Callable
from typing import TypeVar

import pydantic
from canonical import VersionedCipherText
from ckms.core import Keychain
from ckms.core import KeySpecification
from ckms.core import parse_dsn
from ckms.core import parse_spec
from headless.types import IEncrypter

from cbra.core.conf import settings
from cbra.types import IDependant
from cbra.types import VersionedMAC


_keychain: Keychain = Keychain()
T = TypeVar('T')


class ApplicationKeychain(IDependant, IEncrypter[VersionedCipherText]):
    """Provides an interface to perform cryptographic operations with
    a set of well-defined cryptographic keys.
    """
    __module__: str = 'cbra.ext.security'
    keychain: Keychain

    @staticmethod
    def load_spec(dsn: str) -> KeySpecification:
        return parse_spec(parse_dsn(dsn))

    @classmethod
    def __inject__(cls): # pragma: no cover
        return cls.setup

    @classmethod
    async def setup(cls):
        must_configure = False
        if not _keychain.has(settings.STORAGE_ENCRYPTION_KEY)\
        and settings.STORAGE_ENCRYPTION_KEY is not None:
            must_configure = True
            spec = cls.load_spec(settings.STORAGE_ENCRYPTION_KEY)
            _keychain.add(settings.STORAGE_ENCRYPTION_KEY, spec)
        if not _keychain.has(settings.STORAGE_SECURE_INDEX_KEY)\
        and settings.STORAGE_SECURE_INDEX_KEY is not None:
            must_configure = True
            spec = cls.load_spec(settings.STORAGE_SECURE_INDEX_KEY)
            _keychain.add(settings.STORAGE_SECURE_INDEX_KEY, spec)
        if must_configure:
            await _keychain
        return cls(_keychain)

    def __init__(self, keychain: Keychain) -> None:
        self.keychain = keychain

    async def create_hmac(
        self,
        value: bytes | str,
        encoding: str = 'utf-8'
    ) -> VersionedMAC:
        """Create a Hash-based Message Authentication Code (HMAC)
        using the default key.
        """
        if not isinstance(value, bytes):
            value = str.encode(value, encoding=encoding)
        k = self.keychain.get(settings.STORAGE_SECURE_INDEX_KEY)
        if k.dsn is None: # pragma: no cover
            raise ValueError("Can not use a key that is not loaded from a DSN.")
        return VersionedMAC(
            dsn=k.dsn,
            mac=await k.sign(value)
        )

    async def decrypt(
        self,
        ct: VersionedCipherText,
        parser: Callable[[VersionedCipherText, bytes], T] = lambda _, pt: pt
    ) -> T:
        k = self.keychain.get(ct.dsn)
        if k.dsn is None: # pragma: no cover
            raise ValueError("Can not use a key that is not loaded from a DSN.")
        pt = await ct.decrypt(k)
        return parser(ct, pt)

    async def encrypt(
        self,
        pt: bytes | str | pydantic.BaseModel,
        encoding: str = 'utf-8'
    ) -> VersionedCipherText:
        """Encrypt plain text `pt` using the default key."""
        k = self.keychain.get(settings.STORAGE_ENCRYPTION_KEY)
        if k.dsn is None: # pragma: no cover
            raise ValueError("Can not use a key that is not loaded from a DSN.")
        if isinstance(pt, pydantic.BaseModel):
            pt = pt.json()
        if isinstance(pt, str):
            pt = str.encode(pt, encoding=encoding)
        ct = await k.encrypt(pt)
        return VersionedCipherText(
            aad=ct.aad,
            ct=bytes(ct), # type: ignore
            iv=ct.iv,
            dsn=k.dsn,
            tag=ct.tag
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import hashlib

from typing import Any
from .itokensigner import ITokenSigner
from .signable import Signable


class SignableOIDCToken(Signable):
    __module__: str = 'cbra.ext.oauth2.types'
    access_token: str | None = None
    authorization_code: str | None = None
    claims: dict[str, Any]

    async def sign(self, signer: ITokenSigner) -> str:
        # TODO: This must be updated in the ckms module to have a key
        # return the actual hashing algorithm, if any.
        claims = dict(self.claims)
        digestargs: list[Any] = []
        length: int
        try:
            if signer.alg != 'EdDSA':
                hasher = hashlib.new(f'sha{int(signer.alg[2:5])}')
                length = hasher.digest_size
            elif signer.crv == 'Ed25519':
                hasher = hashlib.new('sha512')
                length = hasher.digest_size
            elif signer.crv == 'Ed448':
                hasher = hashlib.new('shake_256')
                length = 114
                digestargs = [length]
            else:
                raise ValueError
        except ValueError:
            raise NotImplementedError(f"Unsupported algorithm: {signer.alg}")
        if self.access_token is not None:
            hasher.update(str.encode(self.access_token))
            digest = hasher.digest(*digestargs)
            claims['at_hash'] = base64.urlsafe_b64encode(digest[:int(length/2)])
        if self.authorization_code is not None:
            hasher.update(str.encode(self.authorization_code))
            digest = hasher.digest(*digestargs)
            claims['c_hash'] = base64.urlsafe_b64encode(digest[:int(length/2)])
        return await signer.jwt(
            claims=claims,
            typ='jwt'
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable
from typing import Callable
from typing import Union

import pydantic

from .icredential import ICredential
from .irequestprincipal import IRequestPrincipal
from .isubject import ISubject
from .nullrequestprincipal import NullRequestPrincipal
from .oidcrequestprincipal import OIDCRequestPrincipal
from .rfc9068requestprincipal import RFC9068RequestPrincipal
from .sessionrequestprincipal import SessionRequestPrincipal


class RequestPrincipal(IRequestPrincipal, pydantic.BaseModel):
    # Note that the order is important here.
    __root__: Union[
        RFC9068RequestPrincipal,
        OIDCRequestPrincipal,
        SessionRequestPrincipal,
        NullRequestPrincipal
    ]

    def get_audience(self) -> set[str]:
        return self.__root__.get_audience()

    def get_credential(self) -> ICredential | None:
        return self.__root__.get_credential()

    def has_audience(self) -> bool:
        return self.__root__.has_audience()

    def is_anonymous(self) -> bool:
        return self.__root__.is_anonymous()

    async def resolve(
        self,
        resolve: Callable[..., Awaitable[ISubject]]
    ) -> ISubject:
        return await resolve(self.__root__)

    def validate_audience(self, allow: set[str]) -> bool:
        return self.__root__.validate_audience(allow)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .iuseronboardingservice import IUserOnboardingService
from .isubjectrepository import ISubjectRepository
from .principaltype import PrincipalType
from .publicidentifier import PublicIdentifier
from .subject import Subject
from .subjectlifecycletype import SubjectLifecycleType


__all__: list[str] = [
    'IUserOnboardingService',
    'ISubjectRepository',
    'PrincipalType',
    'PublicIdentifier',
    'Subject',
    'SubjectLifecycleType',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`ResponseType`."""
import enum


class AuthorizationLifecycle(str, enum.Enum):
    accepted    = 'accepted'
    cancelled   = 'cancelled'
    expired     = 'expired'
    granted     = 'granted'
    refused     = 'refused'
    rejected    = 'rejected'
    requested   = 'requested'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any
from typing import Union

import pydantic
from headless.ext.oauth2.models import TokenResponse

from ...types import AuthorizationRequestIdentifier
from ...types import ClientIdentifier
from ...types import FatalAuthorizationException
from ...types import OIDCProvider
from .downstreamauthorizationstate import DownstreamAuthorizationState
from .loginauthorizationstate import LoginAuthorizationState
from .recoveryauthorizationstate import RecoveryAuthorizationState


class ExternalAuthorizationState(pydantic.BaseModel):
    __root__: Union[
        DownstreamAuthorizationState,
        LoginAuthorizationState,
        RecoveryAuthorizationState
    ]

    @property
    def id(self) -> str:
        return self.__root__.state

    @property
    def client_id(self) -> ClientIdentifier:
        return self.__root__.client_id

    @property
    def kind(self) -> str:
        return self.__root__.kind

    @property
    def request_id(self) -> AuthorizationRequestIdentifier:
        return self.__root__.request_id

    @property
    def return_url(self) -> str | None:
        return self.__root__.return_url

    @classmethod
    def new(
        cls,
        provider: OIDCProvider,
        client_id: str,
        kind: str,
        request_id: str,
        redirect_uri: str,
        **kwargs: Any
    ):
        return cls.parse_obj({
            **kwargs,
            'client_id': client_id,
            'iss': provider.client.metadata.issuer,
            'kind': kind,
            'redirect_uri': redirect_uri,
            'request_id': request_id,
        })

    def set_return_url(self, url: str, **params: str) -> None:
        p: list[str] = list(urllib.parse.urlparse(url)) # type: ignore
        p[4] = urllib.parse.urlencode(
            {**params, **dict(urllib.parse.parse_qsl(p[4]))},
            quote_via=urllib.parse.quote
        )
        self.__root__.return_url = urllib.parse.urlunparse(p)

    async def get_redirect_uri(
        self,
        provider: OIDCProvider,
        callback_uri: str
    ) -> str:
        return await provider.get_redirect_uri(
            redirect_uri=callback_uri,
            state=self.__root__.state,
            nonce=self.__root__.nonce
        )
    
    async def obtain(
        self,
        provider: OIDCProvider,
        code: str,
        url: str,
        iss: str | None = None
    ) -> TokenResponse:
        if provider.client.metadata.authorization_response_iss_parameter_supported\
        and iss != self.__root__.iss:
            raise FatalAuthorizationException("Authorization response from untrusted issuer.")
        token = await provider.obtain(
            code=code,
            state=self.__root__.state,
            redirect_uri=url
        )
        _, oidc = await provider.verify(token, nonce=self.__root__.nonce)
        if oidc is None:
            raise FatalAuthorizationException(
                "The OIDC ID Token returned by the authorization server "
                "was malformed, invalid or otherwise unusable."
            )
        if oidc.email is None:
            raise FatalAuthorizationException(
                "The authorization server did not provide the 'email' "
                "claim."
            )
        return token


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from ckms.utils import b64encode_str


class VersionedMAC(pydantic.BaseModel):
    dsn: str
    mac: str

    @pydantic.validator('mac', pre=True)
    def preprocess_mac(cls, value: bytes | str | None) -> str | None:
        if isinstance(value, bytes):
            value = b64encode_str(value)
        return value



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.


def to_language(locale: str) -> str:
    """Turn a locale name (en_US) into a language name (en-us)."""
    p = locale.find("_")
    if p >= 0:
        return locale[:p].lower() + "-" + locale[p + 1 :].lower()
    else:
        return locale.lower()

def to_locale(language: str):
    """Turn a language name (en-us) into a locale name (en_US)."""
    lang, _, country = language.lower().partition("-")
    if not country:
        return language[:3].lower() + language[3:]
    # A language with > 2 characters after the dash only has its first
    # character after the dash capitalized; e.g. sr-latn becomes sr_Latn.
    # A language with 2 characters after the dash has both characters
    # capitalized; e.g. en-us becomes en_US.
    country, _, tail = country.partition("-")
    country = country.title() if len(country) > 2 else country.upper()
    if tail:
        country += "-" + tail
    return lang + "_" + country


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .applicationclient import ApplicationClient
from .clientstorage import ClientStorage
from .cookiestoredauthorizationrequest import CookieStoredAuthorizationRequest
from .currentissuer import CurrentIssuer
from .downstreamerror import DownstreamError
from .downstreamerror import Error
from .downstreamtokentesponse import DownstreamTokenResponse
from .frontenderror import FrontendError
from .frontendtokenresponse import FrontendTokenResponse
from .localauthorizationrequeststate import LocalAuthorizationRequestState
from .localclientprovider import ClientProvider
from .localclientprovider import LocalClientProvider
from .localopenidprovider import LocalOpenIdProvider
from .localstorageauthorizationrequest import LocalStorageAuthorizationRequest
from .requestaccesstoken import RequestAccessToken
from .requestedgrant import RequestedGrant
from .requestingclient import RequestingClient
from .requestresourceowner import RequestResourceOwner
from .resourceserverclient import RequestedResourceServerClient
from .serverkeychain import ServerKeychain
from .tokenbuilder import TokenBuilder
from .tokensigner import TokenSigner


__all__: list[str] = [
    'ApplicationClient',
    'ClientProvider',
    'ClientStorage',
    'CookieStoredAuthorizationRequest',
    'CurrentIssuer',
    'DownstreamError',
    'DownstreamTokenResponse',
    'FrontendError',
    'FrontendTokenResponse',
    'Error',
    'LocalAuthorizationRequestState',
    'LocalClientProvider',
    'LocalOpenIdProvider',
    'LocalStorageAuthorizationRequest',
    'RequestAccessToken',
    'RequestedGrant',
    'RequestingClient',
    'RequestResourceOwner',
    'RequestedResourceServerClient',
    'ServerKeychain',
    'TokenBuilder',
    'TokenSigner',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .languagecontext import LanguageContext
from .translation import Translation


__all__: list[str] = [
    'LanguageContext',
    'Translation',
]

def gettext(message: str) -> str:
    """Return a string holding the translation of `message` in the currently
    active language, or `message` if no translation was found.
    """
    eol_message = message.replace("\r\n", "\n").replace("\r", "\n")
    if eol_message:
        result = LanguageContext.gettext(eol_message)
    else:
        # Return an empty value of the corresponding type if an empty message
        # is given, instead of metadata, which is the default gettext behavior.
        result = type(message)("")
    return result


def ngettext(singular: str, plural: str, number: int):
    return LanguageContext.ngettext(singular, plural, number)


def override(language: str) -> LanguageContext:
    return LanguageContext(language)


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
#from collections import OrderedDict
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar

import fastapi
from fastapi.concurrency import AsyncExitStack
from fastapi.params import Depends
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import solve_dependencies

#from cbra.types import IDependant
from .container import Container
from .requirement import Requirement


R = TypeVar('R')

_injectables: tuple[type, ...] = (
    Depends,
    Requirement
)

def update_requirements(container: Container, func: Callable[..., Any] | Depends | Any) -> None:
    """Traverse the signature tree of the given function to find
    all :class:`Requirement` instances.
    """
    # TODO: this will completely mess up if multiple Application instances
    # are spawned.
    if not callable(func): return None
    if isinstance(func, Requirement):
        func.add_to_container(container)
    elif isinstance(func, Depends):
        return update_requirements(container, func.dependency)
    try:
        signature = inspect.signature(func) # type: ignore
    except ValueError:
        # No signature to inspect.
        return None
    
    #replaced: list[inspect.Parameter] = []
    for param in signature.parameters.values():
        if isinstance(param.default, Requirement):
            param.default.add_to_container(container)
            if param.default.callable():
                update_requirements(container, param.default.factory)
            continue
        if isinstance(param.default, Depends): # type: ignore
            injectable = param.default.dependency
            if injectable is None:
                # Was declared as f(dependency: Callable = fastapi.Depends())
                injectable = param.annotation
            update_requirements(container, injectable)
        #if inspect.isclass(param.annotation)\
        #and issubclass(param.annotation, IDependant)\
        #and param.default != inspect.Parameter.empty:
        #    replaced.append(param.replace(default=fastapi.Depends(param.annotation.__inject__())))
        #    update_requirements(container, replaced[-1])

    #parameters: OrderedDict[str, inspect.Parameter] = OrderedDict(signature.parameters)
    #for param in replaced:
    #    parameters[param.name] = param
    #func.__signature__ = signature.replace(parameters=parameters.values()) # type: ignore



async def run(
    f: Callable[..., Awaitable[R] | R],
    *,
    container: Container | None = None,
    scope: dict[str, Any] | None = None,
    **kwargs: Any
) -> R:
    container = container or Container.fromsettings()
    async with AsyncExitStack() as stack:
        request: fastapi.Request = fastapi.Request({
            **(scope or {}),
            'fastapi_astack': stack,
            'headers': [],
            'query_string': None,
            'type': 'http',
        })

        update_requirements(container, f)
        dependant = get_dependant(call=f, path='/')
        values, errors, *_ = await solve_dependencies(
            request=request,
            dependant=dependant,
            body=None,
            dependency_overrides_provider=None
        )
        kwargs = {
            **values,
            **kwargs
        }
        if errors:
            raise Exception(errors)
        assert callable(dependant.call)
        result = dependant.call(**kwargs)
        if inspect.isawaitable(result):
            result = await result
        return result


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import fastapi
from canonical import EmailAddress

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.types import SessionRequestPrincipal
from ..models import AuthorizationRequestClient
from ..models import AuthorizationRequest
from ..models import ExternalAuthorizationState
from ..models import ResourceOwner
from ..params import TokenBuilder
from ..types import ResourceOwnerIdentifier
from .base import AuthorizationServerEndpoint


class AuthorizationEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    client: AuthorizationRequestClient # type: ignore
    csrf_protect: bool = True
    metrics: cbra.MetricReporter = cbra.MetricReporter('oauth2')
    name: str = 'oauth2.authorize'
    principal: SessionRequestPrincipal # type: ignore
    path: str = '/authorize'
    status_code: int = 303
    summary: str = 'Authorization Endpoint'

    async def get(
        self,
        params: AuthorizationRequest = AuthorizationRequest.depends()
    ) -> fastapi.Response:
        await self.session
        await params.load(
            client=self.client,
            storage=self.storage,
            session_id=self.session.id
        )

        # First check if the client requires authentication with a
        # downstream provider, since the user will be authenticated
        # any way when it returns here.
        if self.client.requires_downstream() and not params.is_authenticated():
            self.session.pop('ctx')
            return await self.on_downstream_required(params)

        if not self.is_authenticated():
            self.session.pop('ctx')
            return await self.on_login_required(params)

        # Create the resource owner.
        subject = await self.get_subject()
        owner, created = await self.get_or_create_resource_owner(
            client=self.client,
            subject_id=self.session.uid
        )
        if created:
            self.logger.debug(
                "Onboarded subject to client (client: %s, "
                "sub: %s, sector: %s, ppid: %s)",
                self.client.client_id, self.session.uid,
                owner.ppid.sector, owner.ppid.value
            )

        assert self.session.uid
        email = owner.email or params.email
        if email is not None and not self.client.allows_email(email):
            raise NotImplementedError
        assert isinstance(email, EmailAddress) or email is None

        # If the Subject has multiple email addresses and no specific
        # email address was forced by the downstream authentication,
        # an email must be selected. Otherwise, the email address is
        # the email address of the Subject.
        if subject.can_select_email() and email is None:
            return await self.on_select_account(params)
        elif not subject.can_select_email():
            # At this point we may assume that the Subject has at least one
            # email address, but if it hasn't, then some error occurred
            # somehwere else and we logout the user.
            try:
                email = subject.get_email()
            except ValueError:
                return await self.on_login_required(params, logout=True)
            params.set_email(email, prompted=False)

        # Also redirect if it was explicitely requested to select and
        # account.
        if params.must_select_account():
            return await self.on_select_account(params)
        
        if self.client.has_acl() and not self.client.is_member(email):
            self.session.pop('ctx')
            self.logger.debug(
                "Resource Owner was denied access to the Application (email: %s)",
                email
            )
            return await self.on_access_denied(params)

        # TODO: The owner must consent to the audiences
        if params.resources:
            owner.resources |= params.resources
            await self.storage.persist(owner)

        # TODO: For some reason the email can be None here. For now,
        # solve this by deauthenticating the user and redirecting to
        # the login endpoint.
        if email is None:
            return await self.on_login_required(params)

        # Check if all claims required by the authorization request scope are known
        # for this Subject. If not, redirect to a web page where the Resource Owner
        # can provide these claims.
        needs = params.needs(subject) # type: ignore
        if needs:
            return await self.on_update_required(params, needs)

        # If this is an OpenID request, create the claims that will be
        # included in the ID Token so they can be displayed to the user.
        builder = TokenBuilder(self.storage, self.get_issuer(), self.client.impl)
        await params.verify(
            builder=builder,
            request=self.request,
            session=self.session.claims,
            client=self.client,
            owner=owner,
            subject=subject # type: ignore
        )
        await params.persist(self.storage) # type: ignore
        if params.must_consent():
            return await self.on_consent_required(params)

        self.delete_cookies()

        # TODO: Sometime the Subject is not activated after returning from
        # the callback endpoint.
        if not subject.is_active():
            subject.activate()
            await self.persist(subject)

        return params.as_response(client=self.client, iss=self.get_issuer())

    async def on_access_denied(self, params: AuthorizationRequest) -> fastapi.Response:
        """Invoked when the :term:`Resource Owner` is denied access based
        on ACL membership.
        """
        self.delete_cookies(exclude={'oauth2.request'})
        return fastapi.responses.RedirectResponse('/denied')

    async def on_consent_required(
        self,
        params: AuthorizationRequest
    ) -> fastapi.Response:
        """Invoked when the :term:`Resource Owner` must consent to the
        requested scope and/or claims.
        """
        q: dict[str, str] = {'next': params.get_authorize_url(self.request)}
        response = fastapi.responses.RedirectResponse(
            status_code=303,
            url=f'/consent?{urllib.parse.urlencode(q)}'
        )
        response.set_cookie(key='oauth2.request', value=params.id)
        return response

    async def on_downstream_required(
        self,
        params: AuthorizationRequest
    ) -> fastapi.Response:
        """Invoked when the end-user must authenticate with a downstream
        identity provider.
        """
        self.logger.debug(
            "Client requires downstream authentication (client_id: %s)",
            self.client.client_id
        )
        async with self.client.get_provider() as provider:
            state = ExternalAuthorizationState.new(
                provider, self.client.client_id, 'downstream', params.id, 
                params.get_authorize_url(self.request),
            )
            url = await state.get_redirect_uri(
                provider=provider,
                callback_uri=str(self.request.url_for('oauth2.callback')),
            )
        state.set_return_url(settings.OAUTH2_DOWNSTREAM_PROMPT_URL, next=url)
        assert state.return_url is not None
        response = fastapi.responses.RedirectResponse(
            status_code=303,
            url=state.return_url
        )
        response.set_cookie(key='oauth2.request', value=params.id)
        await self.storage.persist(state)
        return response

    async def on_login_required(self, params: AuthorizationRequest, logout: bool = False) -> fastapi.Response:
        """Invoked when the end-user needs to establish an authenticated
        session before proceeding with an authorization request.
        """
        if logout:
            self.session.logout()
            params.add_to_session(self.session)
            assert params.session_id == self.session.id
            await params.persist(self.storage)
        self.delete_cookies(exclude={'oauth2.request'})
        self.logger.debug("The Resource Owner must authenticate to proceed")
        p: dict[str, str] = {
            'next': params.get_authorize_url(self.request),
            'request': params.id
        }
        q = urllib.parse.urlencode(p, quote_via=urllib.parse.quote)
        response = fastapi.responses.RedirectResponse(
            status_code=303,
            url=settings.LOGIN_URL + f'?{q}'
        )
        response.set_cookie(key='oauth2.request', value=params.id)
        return response
    
    async def on_select_account(self, params: AuthorizationRequest) -> fastapi.Response:
        """Invoked when the Resource Owner must select a specific email address."""
        q: dict[str, str] = {
            'next': params.get_authorize_url(self.request),
        }
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=f'/select-account?{urllib.parse.urlencode(q)}'
        )

    async def on_update_required(
        self,
        params: AuthorizationRequest,
        needs: set[str] | None = None
    ) -> fastapi.Response:
        """Invoked when the Resource Owner must update it's profile."""
        q: dict[str, str] = {
            'next': params.get_authorize_url(self.request),
        }
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=f'/update?{urllib.parse.urlencode(q)}'
        )

    @cbra.describe(summary="Authorization Endpoint (OpenID Connect)")
    async def post(self) -> None:
        """The OpenID Connect Core specification mandates that the **Authorization
        Endpoint** must support the HTTP `POST` method. This endpoint takes the
        parameters supported by the `GET` endpoint as the request body, which
        must be provided as `application/json` or `application/x-www-form-urlencoded`.

        *This endpoint is not implemented.*
        """
        # https://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
        raise NotImplementedError
    
    async def get_or_create_resource_owner(
        self,
        client: AuthorizationRequestClient,
        subject_id: int
    ) -> tuple[ResourceOwner, bool]:
        created = False
        owner = await self.storage.get(
            ResourceOwner,
            ResourceOwnerIdentifier(client_id=client.client_id, sub=subject_id)
        )
        if owner is None:
            created = True
            ppid = client.get_pairwise_identifier(subject_id)
            await self.storage.persist(ppid)
            owner = ResourceOwner(
                client_id=client.client_id,
                ppid=ppid,
            )
            await self.storage.persist(owner)
            self.metrics.report('SubjectOnboarded', {
                'client_id': client.client_id,
                'sector_identifier': client.sector_identifier
            })
        return owner, created


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from contextlib import ContextDecorator
from contextvars import ContextVar
from typing import Any

from cbra.core.conf import settings
from .translation import Translation


DEFAULT_LANGUAGE: str = 'en'
catalogs: dict[str | None, Translation] = {
    None: Translation('en')
}
current_language: ContextVar[str] = ContextVar("Language", default=DEFAULT_LANGUAGE)


class LanguageContext(ContextDecorator):
    __module__: str = 'cbra.core.translation'
    _deactivate: bool
    _old_language: str | None = None

    @staticmethod
    def catalog() -> Translation:
        l = LanguageContext.get()
        c = catalogs.get(l)
        if c is None:
            c = catalogs[c] = Translation(l, localedirs=settings.LOCALE_PATHS) # type: ignore
        return c

    @staticmethod
    def get() -> str:
        return current_language.get() or DEFAULT_LANGUAGE

    @staticmethod
    def gettext(message: str) -> str:
        c = LanguageContext.catalog()
        return c.gettext(message)

    @staticmethod
    def ngettext(singular: str, plural: str, number: int) -> str:
        c = LanguageContext.catalog()
        return c.ngettext(singular, plural, number)

    @staticmethod
    def activate(language: str | None) -> None:
        current_language.set(language or DEFAULT_LANGUAGE)

    def __init__(self, language: str | None, deactivate: bool = False):
        self._deactivate = deactivate
        self.language = language

    def __enter__(self):
        self._old_language = self.get()
        if self.language is not None:
            self.activate(self.language)

    def __exit__(self, *args: Any, **kwargs: Any):
        if self._old_language is not None:
            self.activate(self._old_language)



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext.oauth2 import OIDCToken

from cbra.ext.oauth2 import BaseFrontendStorage
from cbra.ext.oauth2.models import ManagedGrant
from cbra.ext.oauth2.types import AuthorizationState
from cbra.ext.oauth2.types import AuthorizationStateIdentifier
from cbra.ext.oauth2.types import ManagedGrantIdentifier
from cbra.ext.oauth2.types import OIDCTokenSubjectIdentifier
from cbra.ext.oauth2.types import ResourceAccessTokenIdentifier
from cbra.ext.oauth2.types import ResourceServerAccessToken
from .basemodelstorage import BaseModelStorage


class ClientStorage(BaseModelStorage, BaseFrontendStorage):
    __module__: str = 'cbra.ext.google.impl.oauth2'

    async def get_grant(self, oid: ManagedGrantIdentifier) -> ManagedGrant | None:
        key = self.key('ManagedGrant', str(oid))
        return self.entity_to_model(
            ManagedGrant,
            await self.get_entity_by_key(key),
            id=str(oid)
        )

    async def persist_access_token(self, obj: ResourceServerAccessToken) -> None:
        entity = self.model_to_entity(
            self.key(
                kind='ResourceServerAccessToken',
                identifier=f'{obj.grant_id}/{obj.resource}',
            ),
            obj,
        )
        await self.put(entity)

    async def get_access_token(
        self,
        oid: ResourceAccessTokenIdentifier
    ) -> ResourceServerAccessToken | None:
        key = self.key(
            kind='ResourceServerAccessToken',
            identifier=f'{oid.grant_id}/{oid.resource}',
        )
        return self.entity_to_model(
            ResourceServerAccessToken,
            await self.get_entity_by_key(key),
            grant_id=oid.grant_id,
            resource=oid.resource
        )

    async def get_authorization_state(self, oid: AuthorizationStateIdentifier) -> AuthorizationState | None:
        key = self.key('AuthorizationState', str(oid))
        return self.entity_to_model(
            AuthorizationState,
            await self.get_entity_by_key(key),
            state=oid
        )

    async def get_oidc_token(self, oid: OIDCTokenSubjectIdentifier) -> OIDCToken | None:
        key = self.key('OIDCToken', str(oid))
        return self.entity_to_model(OIDCToken, await self.get_entity_by_key(key))

    async def persist_authorization_state(self, obj: AuthorizationState) -> None:
        """Persist a :class:`~cbra.ext.oauth2.AuthorizationState` instance."""
        entity = self.model_to_entity(
            self.key('AuthorizationState', str(obj.state)),
            obj,
            exclude={'state'}
        )
        await self.put(entity)

    async def persist_grant(self, obj: ManagedGrant) -> None:
        entity = self.model_to_entity(
            self.key('ManagedGrant', obj.id),
            obj,
            exclude={'id', 'sub'}
        )
        await self.put(entity)

    async def persist_oidc_token(self, obj: OIDCToken) -> None:
        entity = self.model_to_entity(
            self.key('OIDCToken', obj.subject.sha256),
            obj,
            exclude_none=True
        )
        await self.put(entity)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import operator

import pydantic

from cbra.types import IAuthorizationContext
from cbra.types import PolicyPrincipal
from ..types import IPermissionsFinder
from .binding import Binding
from .member import Member


class Policy(pydantic.BaseModel):
    bindings: list[Binding] = pydantic.Field(
        default=[],
        title="Bindings",
        description=(
            "Associates a list of `members`, or principals, with a `role`."
        )
    )

    @property
    def members(self) -> set[Member]:
        return functools.reduce(
            operator.or_,
            [set(x.members) for x in self.bindings]
        )

    async def grants(
        self,
        finder: IPermissionsFinder,
        ctx: IAuthorizationContext
    ) -> set[str]:
        """Return the list of permissions that are granted by this
        policy.
        """
        roles: set[str] = set()
        for binding in self.bindings:
            if binding.is_conditional():
                raise NotImplementedError
            if not binding.is_satisfied(ctx):
                continue
            roles.add(binding.role)
        return await finder.get(roles)

    def add(self, role: str, member: PolicyPrincipal | str) -> None:
        """Add a member to the given binding."""
        if not isinstance(member, PolicyPrincipal):
            member = PolicyPrincipal(member)
        binding = self.get(role)
        if binding is None:
            binding = Binding(role=role)
            self.bindings.append(binding)
        if member not in binding.members:
            binding.members.append(member)

    def get(self, role: str) -> Binding | None:
        """Return the binding for the given role."""
        result = None
        for binding in self.bindings:
            if binding.role != role:
                continue
            result = binding
            break
        return result
    
    def merge(self, policy: 'Policy') -> None:
        """Merges the policies."""
        for updated in policy.bindings:
            binding = self.get(updated.role)
            if binding is None:
                binding = self.update(updated.role, set(updated.members))
                continue
            binding.update(set(updated.members))
    
    def roles(self, member: PolicyPrincipal) -> set[str]:
        return {
            binding.role for binding in self.bindings
            if binding.has(member)
        }
    
    def update(self, role: str, members: set[PolicyPrincipal | str]) -> None:
        """Add a member to the given binding."""
        for member in members:
            self.add(role, member)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable
from typing import Callable
from typing import TypeVar
from typing import Union

import fastapi
import pydantic
from headless.ext.oauth2 import Client
from headless.ext.oauth2 import TokenResponse

from cbra.types import IDependant
from .authorizationstate import AuthorizationState
from .authorizationstateidentifier import AuthorizationStateIdentifier
from .authorizecoderesponse import AuthorizeCodeResponse
from .authorizeerrorresponse import AuthorizeErrorResponse
from .authorizetokenresponse import AuthorizeTokenResponse
from .fatalauthorizationexception import FatalAuthorizationException
from .ifrontendstorage import IFrontendStorage


T = TypeVar('T', bound='AuthorizeResponse')


class AuthorizeResponse(pydantic.BaseModel, IDependant):
    __root__: Union[
        AuthorizeCodeResponse,
        AuthorizeErrorResponse,
        AuthorizeTokenResponse
    ]

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        return cls.parse_query

    @classmethod
    async def parse_query(
        cls: type[T],
        code: str | None = fastapi.Query(
            default=None,
            title='Authorization code',
            description=(
                'The authorization code generated by the authorization server. '
                'Required if using the Authorization Code Flow and the response '
                'mode is `query`, otherwise this parameter is ignored.'
            )
        ),
        state: AuthorizationStateIdentifier | None = fastapi.Query(
            default=None,
            title='State',
            description=(
                'An opaque value used by the client to maintain state '
                'between the request and callback. The authorization '
                'server includes this value when redirecting the '
                'user-agent back to the client. If supplied, this '
                'parameter **must** equal the `state` parameter used when '
                'creating the authorization request. The `state` parameter '
                'is ignored when using JARM because it is included in the '
                'JSON Web Token supplied using the `jwt` parameter per '
                'chosen response mode.'
            )
        ),
        error: str | None = fastapi.Query(
            default=None,
            title="Error code",
            description=(
                "The error code returned by the authorization server if "
                "the user cancelled the request, refused consent, or "
                "failed to authenticate."
            )
        ),
        error_description: str | None = fastapi.Query(
            default=None,
            title="Error description",
            description=(
                "Human-readable ASCII [USASCII] text providing "
                "additional information, used to assist the client "
                "developer in understanding the error that occurred."
            )
        ),
        error_uri: str | None = fastapi.Query(
            default=None,
            title="Error URI",
            description=(
                "A URI identifying a human-readable web page with information "
                "about the error, used to provide the client developer with "
                "additional information about the error."
            )
        ),
        iss: str | None = fastapi.Query(
            default=None,
            title='Issuer',
            description=(
                'Identifies the authorization server that redirected to '
                'this endpoint, as defined in RFC 9207. If the client supports '
                'OAuth 2.0 Authorization Server Issuer Identification, then '
                'this parameter is **required**, if the response mode is not '
                'JWT Secured Authorization Response Mode (JARM) per RFC 9101.\n\n'
                'The `iss` parameter value is the issuer identifier of '
                'the authorization server that created the authorization '
                'response, as defined in RFC 8414.  Its value **must** '
                'be a URL that uses the `https` scheme without any '
                'query or fragment components.'
            )
        ),
        response: str | None = fastapi.Query(
            default=None,
            title="Response",
            alias='response',
            description=(
                "If the response mode is `query.jwt`, then this parameter contains the JSON "
                "Web Token (JWT) containing the response from the authorization server. The "
                "JWT is signed first by the authorization server, then optionally encrypted "
                "using a pre-registered public key that is associated to the client."
            )
        )
    ) -> T:
        if iss:
            raise NotImplementedError("RFC 9207 is not implemented")
        self = cls.parse_obj({
            'code': code,
            'error': error,
            'error_description': error_description,
            'error_uri': error_uri,
            'iss': iss,
            'response': response,
            'state': state
        })
        if isinstance(self.__root__, AuthorizeTokenResponse):
            raise NotImplementedError(
                "JWT Secured Authorization Response Mode is not implemented."
            )
        return self
    
    async def decrypt(self, client: Client) -> None:
        raise NotImplementedError

    async def obtain(self, client: Client, params: AuthorizationState) -> TokenResponse:
        assert isinstance(self.__root__, AuthorizeCodeResponse)
        return await self.__root__.obtain(client, params)

    async def verify(
        self,
        client: Client,
        storage: IFrontendStorage
    ) -> AuthorizationState:
        """Verifies that the authorization response is issued by the authorization
        server and that the response parameters are correct."""
        if isinstance(self.__root__, AuthorizeErrorResponse):
            self.__root__.raise_exception()
            assert False
        if isinstance(self.__root__, AuthorizeTokenResponse):
            await self.decrypt(client)
            assert isinstance(self.__root__, AuthorizeCodeResponse)
        params = await storage.get(self.get_state())
        if params is None:
            raise FatalAuthorizationException(
                "No authorization request satisfied the query parameters."
            )
        return params

    def get_state(self) -> AuthorizationStateIdentifier:
        assert isinstance(self.__root__, AuthorizeCodeResponse)
        return self.__root__.state


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone
from typing import Any

import pydantic
from canonical import AccessCode
from canonical import EmailAddress

from cbra.types import IEmailSender
from cbra.types import PersistedModel


class EmailChallenge(PersistedModel):
    """Maintains information regarding a challenge sent to an email address
    in order to prove the ownership.
    """
    id: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(48),
        primary_key=True
    )

    code: AccessCode = pydantic.Field(
        default=...
    )

    email: EmailAddress = pydantic.Field(
        default=...
    )

    requested: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )

    completed: datetime | None = pydantic.Field(
        default=None
    )

    @classmethod
    def new(cls, email: EmailAddress, **kwargs: Any):
        return cls.parse_obj({
            **kwargs,
            'code': AccessCode.new(length=6, max_attempts=3, ttl=900),
            'email': email
        })

    def age(self) -> int:
        return int((datetime.now(timezone.utc) - self.requested).total_seconds())

    def get_challenge_id(self) -> str:
        return self.id

    def is_blocked(self) -> bool:
        return self.code.is_blocked() or self.code.is_expired()

    def verify(self, code: str) -> bool:
        is_completed = self.code.verify(code)
        if is_completed:
            self.completed = datetime.now(timezone.utc)
        return is_completed
    
    async def send(
        self,
        service: IEmailSender,
        sender: str,
        subject: str,
        template: str = 'email-verification.html.j2',
        context: dict[str, Any] | None = None
    ) -> None:
        await service.send_template(
            sender=sender,
            recipients=[self.email],
            subject=subject,
            template=template,
            context={
                **(context or {}),
                'access_code': self.code.secret,
            }
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast

import pydantic

import cbra.core as cbra
from cbra.types import PersistedModel
from cbra.types import PolicyPrincipal
from cbra.ext.rbac.models import Policy
from cbra.ext.rbac.types import IPolicyManager
from ...basemodelrepository import BaseModelRepository
from ...types import IDatastoreKey


class DatastorePolicyManager(BaseModelRepository, IPolicyManager):
    __module__: str = 'cbra.ext.google.impl.rbac'

    class storage_model(PersistedModel, Policy):
        kind: str
        status: str = pydantic.Field(default='default', primary_key=True)

        class Config:
            title: str = 'IAMPolicy'

    async def attach(self, resource: PersistedModel, policy: Policy):
        """Attaches `policy` to the resource specified by `key`."""
        key = self.model_key(resource)
        await self.persist_model(
            self.storage_model.parse_obj({
                **policy.dict(),
                'kind': key.kind # type: ignore
            }),
            parent=resource
        )

    async def get(
        self,
        model: type[PersistedModel] | PersistedModel,
        id: int | str | None = None
    ) -> Policy:
        """Get the role-based access policy for the given resource."""
        k = self.model_key(self.storage_model, 'default', parent=self.model_key(model, id))
        p = Policy()
        e = await self.get_entity_by_key(k)
        if e is not None:
            p = Policy.parse_obj(e)
        return p
    
    async def get_keys_by_princpal(self, model: type[PersistedModel], principals: set[PolicyPrincipal]) -> set[IDatastoreKey]:
        q = self.query(kind=self.get_entity_name(self.storage_model))
        q.add_filter('kind', '=', self.get_entity_name(model))
        q.add_filter('bindings.members', 'IN', list(principals))

        return {
            cast(IDatastoreKey, self.client.key(*e.key.flat_path[:-2])) # type: ignore
            for e in await self.run_in_executor(q.fetch)
        }


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import DomainName


ALLOWED_DOMAINS: set[DomainName] = {
    DomainName('gmail.com'),
    DomainName('outlook.com'),
    DomainName('hotmail.com'),
    DomainName('live.com'),
    DomainName('myyahoo.com'),
    DomainName('yahoo.com'),
    DomainName('rocketmail.com'),
    DomainName('icloud.com'),
    DomainName('pm.me'),
    DomainName('protonmail.com')
}

DEFAULT_SIGNING_KEY: str = 'oauth2.sig'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from typing import Any
from typing import Literal

import pydantic
from canonical import DomainName
from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken

from cbra.ext.security import ApplicationKeychain
from cbra.types import ISessionManager
from cbra.types import PersistedModel
from cbra.types import VersionedCipherText
from ..types import PrincipalType
from ..types import SubjectLifecycleType
from .principal import Principal
from .subjectclaimset import SubjectClaimSet


class Subject(PersistedModel):
    kind: Literal['User']
    uid: int | None = pydantic.Field(
        default=None,
        auto_assign=True
    )
    created: datetime
    seen: datetime
    status: SubjectLifecycleType = SubjectLifecycleType.pending

    claims: VersionedCipherText | SubjectClaimSet = pydantic.Field(
        default_factory=SubjectClaimSet
    )

    principals: set[Principal] = set()

    #: TODO: The PeristedModel must be refactored to register attribute mutations
    #: so that the storage backend can detect if it needs to delete something.
    _removed_principals: list[Principal] = pydantic.PrivateAttr([])

    def activate(self) -> None:
        self.status = SubjectLifecycleType.active

    def add_principal(
        self,
        issuer: str,
        value: PrincipalType,
        asserted: datetime,
        trust: bool = False
    ) -> None:
        assert self.uid is not None
        new = Principal.new(self.uid, issuer, value, asserted=asserted, trust=trust)
        old = None
        if new in self.principals:
            # TODO: ugly
            principals = list(self.principals)
            old = principals[principals.index(new)]
        must_add = any([
            old is None,
            old is not None and not old.trust
        ])
        if must_add:
            if old is not None:
                self._removed_principals.append(old)
                self.principals.remove(old)
            self.principals.add(new)

    def add_to_session(self, session: ISessionManager[Any]) -> None:
        raise NotImplementedError
    
    def can_destroy(self) -> bool:
        return self.status == SubjectLifecycleType.pending

    def can_select_email(self) -> bool:
        return len([x for x in self.principals if x.spec.kind == 'EmailAddress']) > 1

    def get_claims(self) -> dict[str, Any]:
        assert isinstance(self.claims, SubjectClaimSet)
        return self.claims.dict(exclude_none=True)

    def get_email(self) -> EmailAddress:
        """Return the oldest email adress of the :class:`Subject`. If the
        :class:`Subject` does not have any email addresses, raise 
        :exc:`ValueError`.
        """
        email: EmailAddress | None = None
        for principal in sorted(self.principals, key=lambda x: x.asserted.timestamp()):
            if principal.spec.kind != 'EmailAddress':
                continue
            email = principal.spec.email
            break
        if email is None:
            raise ValueError("Subject does not have any email address.")
        return email

    def get_principals(self) -> list[Principal]:
        return list([x for x in sorted(self.principals, key=lambda x: x.key)])

    def has_claim(self, claim: str) -> bool:
        assert isinstance(self.claims, SubjectClaimSet)
        return getattr(self.claims, claim, None) is not None

    def has_principal(self, principal: PrincipalType) -> bool:
        p = Principal.new(
            0,
            '',
            principal,
            asserted=datetime.now(),
            trust=False
        )
        return p in self.principals

    def is_active(self) -> bool:
        return self.status == SubjectLifecycleType.active

    def is_encypted(self) -> bool:
        return isinstance(self.claims, VersionedCipherText)

    def merge(self, other: 'Subject') -> None:
        assert self.uid is not None
        assert isinstance(self.claims, SubjectClaimSet)
        assert isinstance(other.claims, SubjectClaimSet)
        for principal in other.principals:
            principal.subject = self.uid
            self.principals.add(principal)

    def needs_fallback_email(self, allow: set[DomainName]) -> bool:
        return not any([
            p.spec.email.domain in allow
            for p in self.principals
            if p.spec.kind == 'EmailAddress'
        ])
    
    def update_oidc(self, oidc: OIDCToken) -> None:
        assert not self.is_encypted()
        assert isinstance(self.claims, SubjectClaimSet)
        self.claims = SubjectClaimSet.parse_obj({
            **oidc.dict(exclude_none=True),
            **self.claims.dict(exclude_none=True)
        })

    def update(self, claims: dict[str, Any], updated_at: datetime) -> None:
        """Update the :class:`Subject` with the given `claims`."""
        assert isinstance(self.claims, SubjectClaimSet)
        self.claims = SubjectClaimSet.parse_obj({
            **self.claims.dict(exclude_none=True),
            **claims,
            'updated_at': int(updated_at.timestamp())
        })
    
    async def decrypt(self, keychain: ApplicationKeychain)  -> None:
        assert isinstance(self.claims, VersionedCipherText), self.claims
        self.claims = await keychain.decrypt(
            self.claims,
            parser=lambda _, pt: SubjectClaimSet.parse_raw(pt)
        )
    
    async def encrypt(self, keychain: ApplicationKeychain)  -> None:
        assert not self.is_encypted()
        self.claims = await keychain.encrypt(self.claims)

    def dict(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        # TODO: A very ugly solution for serialization failure.
        try:
            self.principals = list(self.principals) # type: ignore
            return super().dict(*args, **kwargs)
        finally:
            self.principals = set(self.principals)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import inspect
from typing import Any
from typing import Coroutine
from typing import Callable
from typing import TypeVar

from .memorysecret import MemorySecret
from .secret import Secret


__all__: list[str] = ['Secret', 'parse', 'register']

T = TypeVar('T', bound='Secret')

__providers: dict[str, type[Secret]] = {
    'memory': MemorySecret
}


def parse(provider: str, **params: Any) -> Secret:
    if provider not in __providers:
        raise ValueError(f'unknown secret provider: {provider}')
    return __providers[provider].parse_obj({'provider': provider, **params})


@functools.singledispatch
async def load(obj: T) -> T:
    raise TypeError(type(obj).__name__)


def register(provider: str, func: Callable[[T], Coroutine[Any, Any, T]]) -> None:
    sig = inspect.signature(func)
    p = sig.parameters.get('secret')
    if p is None:
        raise TypeError(f'{func.__name__} must accept the \'secret\' parameter.')
    if p.annotation == type(inspect.Parameter.empty())\
    or not inspect.isclass(p.annotation)\
    or not issubclass(p.annotation, Secret):
        raise TypeError(
            "The \'secret\' argument must be annotated with a subclass of "
            "cbra.ext.secrets.Secret"
        )
    load.register(func) # type: ignore
    __providers[provider] = p.annotation


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aorta
import pytest

from cbra.core import MessageRunner


class UnregisteredEvent(aorta.Event):
    pass


class Sewer(aorta.Sewer):
    pass


aorta.register(Sewer)


@pytest.mark.asyncio
async def test_runner_handles_unregistered_event(
    runner: MessageRunner
):
    envelope = UnregisteredEvent().envelope()
    await runner.run(envelope)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class PromptType(str, enum.Enum):
    #: The Authorization Server must not display any authentication or
    #: consent user interface pages. An error is returned if an
    #: End-User is not already authenticated or the Client does not
    #: have pre-configured consent for the requested Claims or does
    #: not fulfill other conditions for processing the request.
    #: The error code will typically be ``login_required``,
    #: ``interaction_required``, or another code defined the OpenID
    #: Connect standard. This can be used as a method to check for
    #: existing authentication and/or consent.
    none = 'none'

    #: The Authorization Server should prompt the End-User for
    #: reauthentication. If it cannot reauthenticate the End-User,
    #: it must return an error, typically ``login_required``.
    login = 'login'

    #: The Authorization Server should prompt the End-User for
    #: consent before returning information to the Client.
    #: If it cannot obtain consent, it must return an error,
    #: typically ``consent_required``.
    consent = 'consent'

    #: The Authorization Server should prompt the End-User to select a user
    #: account. This enables an End-User who has multiple accounts at
    #: the Authorization Server to select amongst the multiple accounts
    #: that they might have current sessions for. If it cannot obtain
    #: an account selection choice made by the End-User, it must return
    #: an error, typically ``account_selection_required``.
    select_account = 'select_account'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Any
from typing import TypeVar

import pydantic
from headless.ext.oauth2.types import GrantType

from ..types import RefreshTokenPolicyType
from ..types import RefreshTokenType


T = TypeVar('T', bound='RefreshToken')


class RefreshToken(pydantic.BaseModel):
    auth_time: int
    created: datetime
    claims: dict[str, Any] = {}
    client_id: str
    granted: datetime
    grant_id: int = 0
    grant_type: GrantType
    expires: datetime
    ppid: int
    renew: RefreshTokenPolicyType
    resources: list[str] = []
    scope: set[str]
    sector_identifier: str
    sub: int
    ttl: int
    token: RefreshTokenType

    @pydantic.root_validator(pre=True) # type: ignore
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        expires = values.get('expires')
        created = values.setdefault('created', datetime.now(timezone.utc))
        ttl = values.get('ttl')
        values.setdefault('granted', created)
        if ttl is not None and not expires:
            assert isinstance(ttl, int)
            values['expires'] = created + timedelta(seconds=ttl)
        if not values.get('token'):
            values['token'] = RefreshTokenType(secrets.token_urlsafe(48))
        return values

    def allows_resource(self, resource: str) -> bool:
        """Return a boolean if the refresh token allows the given resource."""
        return resource in self.resources

    def allows_scope(self, requested: set[str]) -> bool:
        """Return a boolean indicating if the :term:`Refresh Token` allows
        issuing an :term:`Access Token` with the given scope.
        """
        return requested <= self.scope

    def refresh(self: T) -> T:
        """Refresh the token."""
        now = datetime.now(timezone.utc)
        grant = type(self).parse_obj({
            **self.dict(),
            'created': datetime.now(timezone.utc),
            'token': RefreshTokenType(secrets.token_urlsafe(48))
        })
        if grant.renew == RefreshTokenPolicyType.rolling:
            grant.expires = now + timedelta(seconds=self.ttl)
        if grant.renew == RefreshTokenPolicyType.static:
            grant.expires = self.expires
        return grant

    def can_use(self, scope: set[str]) -> bool:
        return self.allows_scope(scope)

    def is_active(self) -> bool:
        return datetime.now(timezone.utc) < self.expires


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from ckms.core import Keychain
from ckms.core import parse_dsn

from cbra.core.conf import settings
from ..const import DEFAULT_SIGNING_KEY


__all__: list[str] = ['ServerKeychain']


keychain: Keychain = Keychain()


async def get():
    if settings.OAUTH2_SIGNING_KEY is None:
        raise TypeError("Declare OAUTH2_SIGNING_KEY in settings or environment.")
    if not keychain.has(DEFAULT_SIGNING_KEY):
        keychain.configure({
            DEFAULT_SIGNING_KEY: parse_dsn(settings.OAUTH2_SIGNING_KEY)
        })
        await keychain
    return keychain


ServerKeychain: Keychain = fastapi.Depends(get)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import hashlib
from typing import Any

import pydantic


class IHashable:
    __module__: str = 'cbra.types'

    def digest(self) -> bytes:
        raise NotImplementedError

    @staticmethod
    def _hash(obj: Any) -> bytes:
        h = hashlib.sha3_256()
        if isinstance(obj, IHashable):
            h.update(obj.digest())
        elif isinstance(obj, str):
            h.update(str.encode(obj, encoding='utf-8'))
        elif isinstance(obj, int):
            h.update(
                int.to_bytes(obj, (obj.bit_length() + 7) // 8, 'big')
            )
        elif isinstance(obj, dict):
            for key in sorted(obj.keys()): # type: ignore
                assert isinstance(key, str)
                h.update(IHashable._hash(key))
                h.update(IHashable._hash(obj[key]))
        elif isinstance(obj, list):
            for item in obj: # type: ignore
                h.update(IHashable._hash(item))
        elif isinstance(obj, datetime.datetime):
            h.update(IHashable._hash(obj.isoformat()))
        elif isinstance(obj, pydantic.BaseModel):
            for k in obj.__fields__.keys():
                h.update(str.encode(k, 'ascii'))
                h.update(IHashable._hash(getattr(obj, 'k')))
        elif obj is None:
            pass
        else:
            raise TypeError(f'Unhashable type: {repr(obj)}')
        return h.digest()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest
import pytest_asyncio
from google.cloud.datastore import Client

from cbra.ext.rbac.models import Policy
from cbra.ext.google.impl.rbac import DatastorePolicyManager
from cbra.ext.google import PolymorphicDatastoreRepository
from cbra.types import PersistedModel
from cbra.types import PolicyPrincipal


class ProtectedResource(PersistedModel):
    id: int | None = pydantic.Field(default=None, primary_key=True)
    name: str


@pytest_asyncio.fixture # type: ignore
async def manager(client: Client):
    manager = DatastorePolicyManager(client=client)
    yield manager
    q = manager.query(kind='IAMPolicy')
    for e in await manager.run_in_executor(q.fetch):
        await manager.delete(e.key)


@pytest_asyncio.fixture # type: ignore
async def resource(repo: PolymorphicDatastoreRepository):
    obj = ProtectedResource(
        id=await repo.auto_increment(ProtectedResource),
        name="Hello world!"
    )
    await repo.persist(obj)
    yield obj
    await repo.delete(repo.model_key(obj))


@pytest.mark.asyncio
async def test_attach_policy(
    manager: DatastorePolicyManager,
    resource: ProtectedResource
):
    p1 = Policy()
    p1.add("owner", "user:foo@example.com")
    await manager.attach(resource, p1)
    p2 = await manager.get(ProtectedResource, resource.id)
    assert p2 is not None
    assert p2.bindings == p1.bindings
    assert len(await manager.get_keys_by_princpal(ProtectedResource, {PolicyPrincipal("user:foo@example.com")})) == 1
    assert len(await manager.get_keys_by_princpal(ProtectedResource, {PolicyPrincipal("user:bar@example.com")})) == 0


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any
from typing import TypeVar

from cbra.types import Conflict
from .persister import Persister
from .resourcemodel import ResourceModel


T = TypeVar('T', bound=ResourceModel)


class Create:
    __module__: str = 'cbra.core'

    async def can_create(self, resource: Any) -> bool:
        """Return a boolean indicating if the resource may be created
        based on semantics. If this method returns ``False``, then
        the request is rejected with the ``409`` status code.
        """
        raise NotImplementedError

    async def create(self, resource: T) -> T:
        if not await self.can_create(resource):
            raise Conflict
        resource = await self.perform_create(resource)
        await self.on_created(resource)
        return resource

    async def perform_create(self, resource: T) -> T:
        self = cast(Persister, self)
        return await self.persist(resource, create=True)

    async def on_created(self, resource: Any):
        pass


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .httpheaderprincipal import HTTPHeaderPrincipal


class OpaqueBearerTokenPrincipal(HTTPHeaderPrincipal):

    @classmethod
    def parse_scheme(
        cls,
        values: dict[str, Any],
        scheme: str,
        value: str
    ) -> dict[str, Any]:
        if scheme != 'bearer':
            raise ValueError('this principal requires the Bearer scheme')
        values.update({'token': value})
        return values


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import jinja2
import pytest

from cbra.core import translation


@pytest.mark.parametrize("language,tmsg", [
    ('nl-NL', "Dit is een vertaalde string!"),
    ('nl', "Dit is een vertaalde string!"),
    ('xx', "This is a translated string!")
])
def test_gettext(language: str, tmsg: str):
    env = jinja2.Environment(
        extensions=['jinja2.ext.i18n']
    )
    env.install_gettext_callables( # type: ignore
        gettext=translation.gettext,
        ngettext=translation.ngettext,
        newstyle=True
    )
    t = env.from_string("{{ gettext('This is a translated string!') }}")
    with translation.override(language):
        assert t.render() == tmsg

    assert t.render() == 'This is a translated string!'


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone

from canonical import EmailAddress

from cbra.core.iam.models import Subject


def test_trusted_principal_does_not_get_overwritten_by_untrusted():
    now = datetime.now(timezone.utc)
    iss = 'https://python-cbra.dev.cochise.io'
    email = EmailAddress('foo@bar.baz')
    subject = Subject(kind='User', uid=1, created=now, seen=now)
    subject.add_principal(iss, email, now, True)
    old = list(subject.principals)[0]
    assert old.trust

    subject.add_principal(iss, email, now, False)
    new = list(subject.principals)[0]
    assert old == new
    assert new.trust


def test_trusted_principal_overwrites_untrusted():
    now = datetime.now(timezone.utc)
    iss = 'https://python-cbra.dev.cochise.io'
    email = EmailAddress('foo@bar.baz')
    subject = Subject(kind='User', uid=1, created=now, seen=now)
    subject.add_principal(iss, email, now, False)
    old = list(subject.principals)[0]
    assert not old.trust

    subject.add_principal(iss, email, now, True)
    new = list(subject.principals)[0]
    assert old == new
    assert new.trust


def test_has_principal():
    now = datetime.now(timezone.utc)
    iss = 'https://python-cbra.dev.cochise.io'
    email = EmailAddress('foo@bar.baz')
    subject = Subject(kind='User', uid=1, created=now, seen=now)
    subject.add_principal(iss, email, now, False)
    assert subject.has_principal(email)


def test_add_email_from_different_issuers():
    now = datetime.now(timezone.utc)
    i1 = "https://accounts.google.com"
    i2 = "https://outlook.microsoft.com"
    email = EmailAddress('foo@bar.baz')
    subject = Subject(kind='User', uid=1, created=now, seen=now)
    subject.add_principal(i1, email, now, False)
    subject.add_principal(i2, email, now, False)
    assert len(subject.principals) == 1


def test_override_email_with_trusted_issuer():
    now = datetime.now(timezone.utc)
    i1 = "https://accounts.google.com"
    i2 = "https://outlook.microsoft.com"
    email = EmailAddress('foo@bar.baz')
    subject = Subject(kind='User', uid=1, created=now, seen=now)
    subject.add_principal(i1, email, now, False)
    subject.add_principal(i2, email, now, True)
    assert len(subject.principals) == 1
    assert subject.principals.pop().spec.iss == i2


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pathlib

import pytest

from cbra.core.translation import Translation


@pytest.mark.parametrize("language,tmsg", [
    ("nl_NL", "Hallo wereld!"),
    ("nl", "Hallo wereld!"),
    ("en", "Hello world!"),
    ("de", "Hello world!"),
])
def test_gettext(language: str, tmsg: str):
    curdir = pathlib.Path(__file__).parent
    translations = Translation(language, localedirs=[
        "res/locale",
        curdir.joinpath("locale")
    ])
    assert translations.gettext("Hello world!") == tmsg


@pytest.mark.parametrize("language,tmsg", [
    ("nl_NL", "Dit is een vertaalde string!"),
    ("nl", "Dit is een vertaalde string!"),
])
def test_gettext_merged(language: str, tmsg: str):
    curdir = pathlib.Path(__file__).parent
    translations = Translation(language, localedirs=[
        "res/locale",
        curdir.joinpath("locale")
    ])
    assert translations.gettext("This is a translated string!") == tmsg


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import logging
import os
from types import EllipsisType
from typing import cast
from typing import Any

import sendgrid
from sendgrid.helpers.mail import Attachment
from sendgrid.helpers.mail import Content
from sendgrid.helpers.mail import ContentId
from sendgrid.helpers.mail import Disposition
from sendgrid.helpers.mail import Email
from sendgrid.helpers.mail import FileContent
from sendgrid.helpers.mail import FileName
from sendgrid.helpers.mail import FileType
from sendgrid.helpers.mail import Mail
from sendgrid.helpers.mail import Subject
from sendgrid.helpers.mail import To

from cbra.core.conf import settings
from cbra.types import ServiceNotAvailable
from cbra.ext.email import BaseEmailSender


class SendgridEmailSender(BaseEmailSender):
    __module__: str = 'cbra.ext.sendgrid'
    logger: logging.Logger = logging.getLogger('cbra.endpoint')

    @property
    def sendgrid_api_key(self) -> str:
        if not os.getenv('SENDGRID_API_KEY'):
            self.logger.critical("SENDGRID_API_KEY not specified in environment")
            raise ServiceNotAvailable
        return cast(str, os.getenv('SENDGRID_API_KEY'))

    @property
    def client(self) -> sendgrid.SendGridAPIClient:
        c =  sendgrid.SendGridAPIClient(
            api_key=self.sendgrid_api_key)
        c.client.timeout = 10 # type: ignore
        return c

    async def send_template(
        self,
        sender: str | EllipsisType,
        recipients: list[str],
        subject: str,
        template: str | None = None,
        templates: dict[str, str] | None = None,
        context: dict[str, Any] | None = None,
        content_type: str = "text/html",
        headers: dict[str, Any] | None = None,
        cc: list[str] | None = None,
        bcc: list[str] | None = None,
        attachments: list[dict[str, Any]] | None = None
    ) -> None:
        """Render a template and send it as email to the specified
        recipients.
        """
        if not bool(template) ^ bool(templates):
            raise TypeError("Provide the template or templates parameter.")
        if template:
            templates = {content_type: template}

        assert isinstance(templates, dict)
        ctx: dict[str, Any] = context or {}
        if sender == ...:
            if settings.EMAIL_DEFAULT_SENDER is None:
                raise TypeError('Declare EMAIL_DEFAULT_SENDER in settings')
            sender = settings.EMAIL_DEFAULT_SENDER
        msg = Mail()
        msg.to = [To(x) for x in recipients]
        msg.from_email = Email(sender)
        msg.subject = Subject(subject)
        content: list[Content] = []
        for content_type, body in self.render_templates(templates, ctx):
            content.append(Content(content_type, body))
        msg.content = content
        for params in (attachments or []):
            attachment = Attachment()
            attachment.content_id = ContentId(params['content_id'])
            attachment.file_type = FileType(params['content_type'])
            attachment.file_content = FileContent(params['content'])
            attachment.file_name = FileName(params['filename'])
            attachment.disposition = Disposition('attachment')
            msg.add_attachment(attachment)

        await self._send(msg)

    async def _send(self, msg: Mail) -> bool:
        loop = asyncio.get_running_loop()
        f = self.client.send # type: ignore
        from python_http_client.exceptions import HTTPError

        try:
            response = await loop.run_in_executor( # type: ignore
                None, lambda: f(msg)) # type: ignore
        except HTTPError as e:
            raise Exception(e.to_dict)
        except Exception as e: # pylint: disable=broad-except
            if not hasattr(e, 'status_code'):
                raise
            response = e
            self.logger.critical(
                "Failed to send email (response: %s)",
                response.status_code # type: ignore
            )
            raise ServiceNotAvailable
        if response.status_code >= 400: # type: ignore
            self.logger.critical(
                "Failed to send email (response: %s)",
                response.status_code # type: ignore
            )
            raise ServiceNotAvailable

        return True


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol
from typing import TypeAlias
from typing import Union

from headless.ext.oauth2.models import ManagedGrant

from cbra.ext.security import ApplicationKeychain
from .globalcredentialidentifier import GlobalCredentialIdentifier
from .subjectcredentialidentifier import SubjectCredentialIdentifier


CredentialIdentifierType: TypeAlias = Union[
    GlobalCredentialIdentifier,
    SubjectCredentialIdentifier
]


class IPersistableCredential(Protocol):
    credential: ManagedGrant

    def get_primary_key(self) -> CredentialIdentifierType: ...
    async def persist(self) -> None: ...
    async def decrypt(self, keychain: ApplicationKeychain): ...
    async def encrypt(self, keychain: ApplicationKeychain): ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import NoReturn

import pydantic

from .authorizationstate import AuthorizationState
from cbra.types import IDependant


class AuthorizeErrorResponse(pydantic.BaseModel, IDependant):
    error: str = pydantic.Field(
        default=...
    )

    error_description: str | None = pydantic.Field(
        default=None
    )

    error_uri: str | None = pydantic.Field(
        default=None
    )

    state: AuthorizationState = pydantic.Field(
        default=...
    )

    def raise_exception(self) -> NoReturn:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import ICache
from .cacheconfiguration import CacheConfiguration
from .memorycache import MemoryCache


__all__: list[str] = [
    'CacheConfiguration',
    'MemoryCache'
]


def cache(name: str) -> ICache:
    config = CacheConfiguration.get(name)
    async def f(impl: ICache = fastapi.Depends(config.factory)) -> ICache:
        await impl.connect(config)
        return impl
    return fastapi.Depends(f)





# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Literal

from headless.ext.oauth2.models import TokenResponse
from headless.ext.oauth2.types import GrantType

from ..types import IAuthorizationServerStorage
from ..types import ITokenBuilder
from ..types import ITokenSigner
from ..types import InvalidGrant
from ..types import InvalidTarget
from ..types import RefreshTokenIdentifier
from ..types import RequestedScope
from .basetokenrequest import BaseTokenRequest
from .resourceowner import ResourceOwner


logger: logging.Logger = logging.getLogger('cbra.endpoint')


class RefreshTokenGrant(BaseTokenRequest):
    grant_type: Literal[GrantType.refresh_token]
    owner: ResourceOwner | None
    refresh_token: RefreshTokenIdentifier
    resource: str | None
    scope: set[str] = set()

    async def create_response(
        self,
        storage: IAuthorizationServerStorage,
        builder: ITokenBuilder,
        signer: ITokenSigner
    ) -> TokenResponse:
        grant = await storage.fetch(self.refresh_token)
        if grant is None or not grant.is_active():
            raise InvalidGrant("The refresh token is not valid.")
        if not grant.allows_scope(self.scope):
            raise InvalidGrant("The scope requested exceeds the granted scope.")
        if self.resource and not grant.allows_resource(self.resource):
            raise InvalidTarget(
                "The requested audience is not allowed with this refresh token."
            )
        refreshed = grant.refresh()
        assert refreshed.token != grant.token
        await storage.destroy(grant)
        await storage.persist(refreshed)

        at, expires_in = builder.rfc9068(
            sub=grant.ppid,
            scope=[RequestedScope.parse_obj({'name': s}) for s in set(self.scope or grant.scope)],
            auth_time=grant.auth_time,
            audience=self.resource
        )
        access_token = await at.sign(signer)
        await self.register_issued_access_token(
            token=at,
            signed=access_token,
            scope=list(self.scope or grant.scope),
            storage=storage,
            sub=grant.sub,
            claims=grant.claims
        )
        return TokenResponse(
            access_token=access_token, # type: ignore
            expires_in=expires_in,
            refresh_token=str(refreshed.token),
            token_type='Bearer'
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from google.cloud.datastore import Client

from cbra.core.conf import settings
from cbra.core.ioc import Container
from .environ import GOOGLE_DATASTORE_NAMESPACE
from .environ import GOOGLE_SERVICE_PROJECT


__all__: list[str] = ['setup']


def setup(container: Container) -> None:
    if GOOGLE_SERVICE_PROJECT and GOOGLE_DATASTORE_NAMESPACE:
        container.inject(
            'GoogleDatastoreClient',
            Client(project=GOOGLE_SERVICE_PROJECT, namespace=GOOGLE_DATASTORE_NAMESPACE)
        )
        container.provide('CredentialRepository', {
            'qualname': (
                settings.CREDENTIALS_REPOSITORY or
                'cbra.ext.google.impl.credentials.DatastoreCredentialRepository'
            )
        })
        container.provide('SubjectResolver', {
            'qualname': 'cbra.ext.google.DatastoreSubjectResolver'
        })
        container.provide('SubjectRepository', {
            'qualname': 'cbra.ext.google.DatastoreSubjectRepository'
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from canonical import EmailAddress


class Mailbox(pydantic.BaseModel):
    principal_id: str = pydantic.Field(
        default=...,
        title="Principal ID",
        description="The principal identifying the mailbox."
    )

    email: EmailAddress = pydantic.Field(
        default=...,
        title="Email",
        description=(
            "An email address that a Subject may use to sign in."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any

import fastapi
from aorta.types import Envelope

from .messagepublisher import MessagePublisher
from .messagerunner import MessageRunner


class LocalMessageTransport:
    __module__: str = 'aorta'
    logger: logging.Logger = logging.getLogger('cbra.endpoint')
    runner: MessageRunner

    def __init__(
        self,
        request: fastapi.Request
    ):
        self.publisher = MessagePublisher(request=request, transport=self)
        self.runner = MessageRunner(request=request, publisher=self.publisher)

    async def send(
        self,
        messages: list[Envelope[Any]],
        is_retry: bool = False
    ) -> None:
        for message in messages:
            await self.runner.run(message)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from canonical import DomainName


class ClientInfo(pydantic.BaseModel):
    allowed_email_domains: set[DomainName] = set()
    id: str
    client_uri: str | None = None
    contacts: list[str] = []
    logo_uri: str | None = None
    name: str
    organization_name: str
    policy_uri: str | None = None
    sector_identifier: str
    tos_uri: str | None = None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import overload
from typing import Any
from typing import Protocol

from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken
from headless.ext.oauth2.models import SubjectIdentifier

from .subject import Subject
from .principaltype import PrincipalType


class IUserOnboardingService(Protocol):

    @overload
    async def can_use(
        self,
        subject: Subject,
        principals: list[EmailAddress | SubjectIdentifier]
    ) -> bool: ...

    @overload
    async def can_use(self, token: OIDCToken) -> bool: ...

    @overload
    async def get(self, oidc: OIDCToken) -> Subject | None: ...

    @overload
    async def get(self, uid: int) -> Subject | None: ...

    async def update(
        self,
        subject: Subject,
        iss: str,
        principals: list[Any],
        trust: bool
    ) -> None: ...

    async def update_oidc(
        self,
        subject: Subject,
        oidc: OIDCToken
    ) -> None: ...

    def initialize(self) -> Subject: ...
    async def destroy(self, uid: int) -> None: ...
    async def email(self, issuer: str, email: EmailAddress, trust: bool = False) -> tuple[Subject, bool]: ...
    async def oidc(self, token: OIDCToken) -> tuple[Subject, bool]: ...
    async def persist(self, subject: Subject) -> None: ...
    async def sync(self, issuer: str, principal: PrincipalType) -> tuple[Subject, bool]: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from headless.ext.oauth2.models import OIDCToken

from cbra.core.conf import settings
from cbra.ext.credentials import CredentialRepository
from cbra.ext.credentials import ICredentialRepository
from cbra.ext.credentials.types import SubjectCredentialIdentifier
from cbra.types import NotFound
from cbra.types import SessionRequestPrincipal
from .tokenhandlerendpoint import TokenHandlerEndpoint


class FrontendUserEndpoint(TokenHandlerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    credentials: ICredentialRepository = CredentialRepository
    grant_id: SubjectCredentialIdentifier | None = fastapi.Cookie(
        default=None,
        title="Grant ID",
        alias='bff.grant',
        description=(
            "Identifies the grant that is used to exchange tokens with "
            "the authorization server."
        )
    )
    name: str = 'bff.user'
    path: str = '/oauth/v2/user'
    principal: SessionRequestPrincipal # type: ignore
    response_model_exclude: set[str] = {
        "aud",
        "azp",
        "at_hash",
        "c_hash",
        "exp",
        "iat",
        "iss",
        "nbf",
        "nonce",
    }
    response_model_exclude_none: bool = True
    summary: str = "Current User"

    async def get(self) -> OIDCToken:
        await self.session
        if not self.session.subject or not self.grant_id:
            self.logger.debug("User is not authenticated")
            raise NotFound
        dto = await self.credentials.get(self.grant_id)
        if dto is None:
            self.logger.debug("No OIDC token found for authenticated user.")
            raise NotFound
        oidc = dto.credential.oidc
        if oidc.email:
            oidc.email_verified = settings.APP_ISSUER_TRUST # TODO
        return oidc
    
    async def delete(self) -> None:
        await self.session
        self.session.logout()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
from starlette.datastructures import URL

from cbra.types import IQueryResult
from .const import CURSOR_PARAM_NAME
from .resourcemodel import ResourceModel


class Collection:
    __module__: str = 'cbra.core'
    _limit: int
    cursor_param_name: str
    default_limit: int = 100
    max_limit: int
    model: type[ResourceModel]

    async def list(
        self,
        page_token: str | None = fastapi.Query(
            default=None,
            alias=CURSOR_PARAM_NAME,
            title="Pagination token",
            description="Pagination token used to iterate over the collection."
        ),
        limit: int | None = fastapi.Query(
            default=None,
            title="Limit",
            description=(
                "Limits the number of resources included in the result set. "
                "Note that the resource server may impose an arbitrary "
                "limit on the size of the result set, regardless of the "
                "client-provided value of `limit`."
            )
        )
    ) -> Any:
        self._limit = min(self.max_limit, limit or self.default_limit)
        result = await self.filter(
            None,
            page_token=page_token,
            limit=self._limit
        )
        resource =  self.model.__list_model__.parse_obj({
            'apiVersion': 'v1',
            'kind': f'{self.model.__name__}List',
            'metadata': {
                'nextUrl': self.get_next_url(result.token)
            },
            'items': []
        })
        resource.items = [self.adapt(x) for x in result.items] # type: ignore
        return resource

    async def filter(
        self,
        params: Any,
        page_token: str | None,
        limit: int
    ) -> IQueryResult[Any]:
        raise NotImplementedError
    
    def get_next_url(self, token: str | None) -> str | None:
        if not token:
            return None
        url: URL = self.reverse('list', params={self.cursor_param_name: token}) # type: ignore
        url = url.include_query_params( # type: ignore
            limit=self._limit
        )
        return str(url) # type: ignore


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Generator
from typing import TypeVar

from pydantic.validators import str_validator


T = TypeVar('T', bound='ETagSet')


class ETagSet(str):
    __module__: str = 'cbra.types'
    openapi_format: str = 'comma-separated'
    openapi_type: str = 'string'

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None:
        field_schema.update( # pragma: no cover
            type=cls.openapi_format,
            format=cls.openapi_format
        )

    @classmethod
    def __get_validators__(cls: type[T]) -> Generator[Callable[..., str | T], None, None]:
        yield str_validator

    def toset(self) -> set[str]:
        return set(filter(bool, [str.strip(x) for x in self.split(',')]))


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Tests that a model field that is declared as optional is also allowed
on the response model.
"""
import pydantic
import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra


class Model(cbra.ResourceModel):
    id: int = pydantic.Field(..., primary_key=True)
    value: str | None


class NoneResource(cbra.Resource, model=Model):

    async def retrieve(self, model_id: int) -> Model:
        return Model(id=1, value=None)


@pytest_asyncio.fixture(scope='session') # type: ignore
async def client():
    app = cbra.Application()
    app.add(NoneResource)
    async with httpx.Client(base_url='https://cbra', app=app) as client:
        yield client



@pytest.mark.asyncio
async def test_returns_200(
    client: httpx.Client
):
    response = await client.get(url='/models/1')
    assert response.status_code == 200, response.content


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest
import pytest_asyncio
from google.cloud.datastore import Client

from cbra.ext.google import TypedDatastoreRepository
from cbra.types import PersistedModel
from cbra.types import VersionConflict


class TypedEntity(PersistedModel):
    id: int = pydantic.Field(default=..., primary_key=True)
    name: str


class TypedEntityRepository(TypedDatastoreRepository[TypedEntity]):
    model = TypedEntity


@pytest.fixture
def repo(client: Client):
    return TypedEntityRepository(client=client)


@pytest_asyncio.fixture # type: ignore
async def resource(repo: TypedEntityRepository):
    obj = TypedEntity(
        id=await repo.allocate_pk(),
        name="Hello world!"
    )
    await repo.persist(obj)
    yield obj
    await repo.delete(repo.model_key(obj))


@pytest.mark.asyncio
async def test_find_by_pk(
    repo: TypedEntityRepository,
    resource: TypedEntity
):
    obj = await repo.get(resource.id)
    assert obj is not None
    assert obj.id == resource.id
    assert obj.name == resource.name


@pytest.mark.asyncio
async def test_find_by_filters(
    repo: TypedEntityRepository,
    resource: TypedEntity
):
    obj = await repo.one([('name', '=', 'Hello world!')])
    assert obj is not None
    assert obj.id == resource.id
    assert obj.name == resource.name


@pytest.mark.asyncio
async def test_persist(
    repo: TypedEntityRepository,
    resource: TypedEntity
):
    o1 = await repo.get(resource.id)
    assert o1 is not None

    o1.name = "Foo"
    await repo.persist(o1)

    o2 = await repo.get(resource.id)
    assert o2 is not None
    assert o1.id == o2.id
    assert o2.name == "Foo"



@pytest.mark.asyncio
async def test_persist_concurrent(
    repo: TypedEntityRepository,
    resource: TypedEntity
):
    assert resource.__metadata__.generation == 1
    o1 = await repo.get(resource.id)
    assert o1 is not None
    assert o1.__metadata__.generation == 1

    o1.name = "Foo"
    assert o1 == await repo.persist(o1)
    assert o1.__metadata__.generation == 2
    assert resource.__metadata__.generation == 1

    resource.name = "Bar"
    with pytest.raises(VersionConflict):
        await repo.persist(resource)

    o1.name = "Baz"
    await repo.persist(o1)
    assert o1.__metadata__.generation == 3


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import Callable
from typing import Union
from typing import TypeVar

import aorta
import fastapi
from fastapi.routing import DecoratedCallable
from fastapi.params import Depends

from .endpoint import Endpoint
from .ioc import Container
from .ioc import Requirement
from .resource import Resource


T = TypeVar('T', bound='EndpointProvider')


class EndpointProvider:
    container: Container
    _injectables: tuple[type, ...] = (
        Depends,
        Requirement
    )

    def __init__(self):
        self.container = Container.fromsettings()

    def add(
        self: T,
        routable: Union[
            type[Endpoint | Resource],
            type[aorta.EventListener |aorta.CommandHandler],
            type[fastapi.APIRouter],
            fastapi.APIRouter,
            T,
            Any
        ],
        *args: Any, **kwargs: Any
    ) -> None:
        if hasattr(routable, 'add_to_router'):
            routable.add_to_router(self, *args, **kwargs) # type: ignore
        elif inspect.isclass(routable) and issubclass(routable, (Endpoint, Resource)):
            routable.add_to_router(self, *args, **kwargs) # type: ignore
        elif inspect.isclass(routable) and issubclass(routable, aorta.MessageHandler):
            # Ensure that all members are injected in the container.
            for _, member in inspect.getmembers(routable):
                if not isinstance(member, (Depends, Requirement)):
                    continue
                self.update_requirements(member)
            aorta.register(routable)
        elif inspect.isclass(routable) and issubclass(routable, type(self)):
            self.add(routable(), *args, **kwargs)
        elif inspect.isclass(routable) and issubclass(routable, fastapi.APIRouter):
            self.add(routable(), *args, **kwargs)
        elif isinstance(routable, fastapi.APIRouter):
            self.include_router(routable, *args, **kwargs) # type: ignore
        else:
            raise NotImplementedError

    def discover_requirements(
        self,
        decorator_factory: Callable[..., DecoratedCallable],
        *args: Any,
        **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.update_requirements(func)
            return decorator_factory(self, *args, **kwargs)(func)
        return decorator

    def update_requirements(self, func: Callable[..., Any] | Depends | Any) -> None:
        """Traverse the signature tree of the given function to find
        all :class:`Requirement` instances.
        """
        # TODO: this will completely mess up if multiple Application instances
        # are spawned.
        if isinstance(func, Requirement):
            func.add_to_container(self.container)
        elif isinstance(func, Depends):
            return self.update_requirements(func.dependency)
        if not callable(func):
            return None
        try:
            signature = inspect.signature(func) # type: ignore
        except ValueError:
            # No signature to inspect.
            return None
        for param in signature.parameters.values():
            if not isinstance(param.default, self._injectables):
                continue
            if isinstance(param.default, Requirement):
                param.default.add_to_container(self.container)
                if param.default.callable():
                    self.update_requirements(param.default.factory) # type: ignore
                continue

            if isinstance(param.default, Depends):
                injectable = param.default.dependency
                if injectable is None:
                    # Was declared as f(dependency: Callable = fastapi.Depends())
                    injectable = param.annotation
                self.update_requirements(injectable)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .accesstokenfactory import AccessTokenFactory
from .endpoints import BaseAuthorizationEndpoint
from .endpoints import BaseCallbackEndpoint
from .models import ManagedGrant


__all__: list[str] = [
    'AccessTokenFactory',
    'BaseAuthorizationEndpoint',
    'BaseCallbackEndpoint',
    'ManagedGrant'
]


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any

import aorta

import cbra.core as cbra
from cbra.core import MessageRunner
from cbra.types import CommandFailed
from .eventarcendpoint import EventarcEndpoint
from .messagediscarded import MessageDiscarded
from .pubsubmessage import PubsubMessage


class AortaEndpoint(EventarcEndpoint):
    __module__: str = 'cbra.ext.google'
    runner: MessageRunner = MessageRunner.depends()

    async def put(self, dto: dict[str, Any]):
        envelope = aorta.parse(dto)
        if envelope is None:
            self.logger.warning("Could not parse the request body as a known Aorta message type.")
            return
        self.logger.debug("Received message %s", envelope.kind)

    async def run_command(self, envelope: aorta.types.Envelope[Any]):
        success, *_ = await self.runner.run(envelope)
        if not success:
            self.logger.critical(
                "Refusing to acknowledge message (id: %s, kind: %s)",
                envelope.metadata.uid, envelope.kind
            )
            raise CommandFailed

    async def on_message(self, message: PubsubMessage) -> None:
        try:
            data = message.get_data()
        except ValueError:
            self.logger.critical("Data could not be interpreted as JSON.")
            raise MessageDiscarded
        if data is None:
            raise MessageDiscarded
        envelope = aorta.parse(data)
        if envelope is None:
            self.logger.critical("Message is not an Aorta message type.")
            raise MessageDiscarded
        envelope = cast(aorta.types.Envelope[Any], envelope)
        if envelope.is_command():
            await self.run_command(envelope)
        elif envelope.is_event():
            _, retry = await self.runner.run(envelope)
            for m in retry:
                self.logger.critical(
                    "Handler failure (handler: %s, message: %s)",
                    m.metadata.handler, m.metadata.uid
                )
        else:
            self.logger.warning(
                "Received Aorta message of unknown type (id: %s, type: %s)",
                envelope.metadata.uid, envelope.type
            )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
import pathlib
from typing import Any
from typing import TypeVar

import jinja2
import yaml

from cbra.core.conf import settings
from cbra.types import Request
from .basestorage import BaseStorage
from .models import AuthorizationServerModel
from .models import Client
from .types import ClientIdentifier
from .types import ObjectIdentifier
from .types import PrincipalIdentifier


T = TypeVar('T', bound=AuthorizationServerModel)


class ApplicationStorage(BaseStorage):
    """A :class:`~cbra.ext.oauth2.BaseStorage` implementation that
    retrieves specific resources from either environment variables,
    local configuration files, or the :mod:`~cbra.core.conf.settings`
    module.
    """
    __module__: str = 'cbra.ext.oauth2'
    clients: dict[str, pathlib.Path | Client]
    client_config_dir: pathlib.Path
    template: jinja2.Environment = jinja2.Environment(
        variable_start_string='${',
        variable_end_string='}'
    )

    def __init__(
        self,
        request: Request,
    ) -> None:
        self.client_config_dir = pathlib.Path(settings.OAUTH2_CLIENT_CONFIG_DIR)
        self.clients = {
            x.with_suffix('').name: x
            for x in self.client_config_dir.glob('**/*.yml')
        }

    async def fetch(self, oid: ObjectIdentifier[T]) -> T | None:
        if isinstance(oid, ClientIdentifier):
            return await super().get(Client, str(oid)) # type: ignore
        else:
            return None

    async def get(self, cls: type[T], *args: Any, **kwargs: Any) -> T | None:
        if cls == Client:
            return await super().get(cls, *args, **kwargs)
        else:
            return None
        
    async def get_client(self, client_id: str) -> Client | None:
        client = self.clients.get(client_id)
        if client is not None and not isinstance(client, Client):
            assert isinstance(client, pathlib.Path)
            with open(client) as f:
                client = await self.load_client_config_file(client_id, f.read())
        return client

    async def load_client_config_file(self, client_id: str, content: str) -> Client:
        t = self.template.from_string(content)
        c = {'env': os.environ, 'settings': settings}
        return Client.parse_obj({
            **yaml.safe_load(t.render(c)), # type: ignore
            'client_id': client_id
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Protocol
from typing import TypeVar

import pydantic

from .ipolymorphiccursor import IPolymorphicCursor
from .ipolymorphicquery import IPolymorphicQuery
from .persistedmodel import PersistedModel
from .policyprincipal import PolicyPrincipal


T = TypeVar('T', bound=PersistedModel)
M = TypeVar('M', bound=pydantic.BaseModel)


class IPolymorphicRepository(Protocol):
    __module__: str = 'cbra.types'

    async def auto_increment(self, obj: type[PersistedModel] | PersistedModel) -> int:
        """Return an auto incrementing integer to be used as a technical
        primary key.
        """
        ...

    async def access(
        self,
        cls: type[T],
        principals: set[PolicyPrincipal],
        ordering: list[str] | None = None,
        limit: int = 100,
        token: str | None = None,
        adapter: Callable[[T], M] | None = None
    ) -> IPolymorphicCursor[M]:
        """Return the objects that the principals have at least one
        grant on.
        """
        ...

    async def allocate(self, name: str) -> int:
        ...

    async def allocate_many(self, name: str, n: int) -> list[int]:
        ...

    async def find(
        self,
        cls: type[T],
        filters: list[tuple[str, str, Any]],
        ordering: list[str] | None = None,
    ) -> None | T:
        ...

    async def list(
        self,
        cls: type[T],
        ordering: list[str] | None = None,
        limit: int = 100,
        token: str | None = None,
        adapter: Callable[[T], M] | None = None
    ) -> IPolymorphicCursor[M]:
        ...
    async def query(self, cls: type[T]) -> IPolymorphicQuery[T]: ...
    async def get(self, cls: type[T], pk: Any) -> None | T: ...
    async def persist(self, obj: T) -> T: ...
    async def persist_many(self, obj: Iterable[T]) -> Iterable[T]: ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any

import ckms.utils
import pydantic
from ckms.core import KeySpecification
from ckms.types import CipherText


DEFAULT_CHUNK_SIZE: int = 255


class VersionedCipherText(pydantic.BaseModel):
    aad: str | None = None
    ct: str | list[str]
    dsn: str
    enc: str | None = None
    iv: str | None = None
    tag: str | None = None

    @pydantic.root_validator(pre=True) # type: ignore
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        ct: bytes | str | list[str] | None = values.get('ct')
        if isinstance(ct, (bytes, str)):
            # Split the cipher text up in chunks because some storage
            # backends have limits on the size of a value. This excludes
            # systems that do no support an array data type, but otherwise
            # we cannot predict if the storage system is able to store the
            # value.
            if isinstance(ct, bytes):
                ct = ckms.utils.b64encode_str(ct)
            values.update({
                'ct': [
                    ct[i:i+DEFAULT_CHUNK_SIZE]
                    for i in range(0, len(ct), DEFAULT_CHUNK_SIZE)
                ]
            })
            assert str.join('', values['ct']) == ct
        return values

    @pydantic.validator('iv', 'aad', 'tag', pre=True) # type: ignore
    def preprocess_bytes(cls, value: bytes | str | None) -> str | None:
        if isinstance(value, bytes):
            value = ckms.utils.b64encode_str(value)
        return value

    def get_ciphertext(self) -> bytes:
        ct = self.ct
        if not isinstance(ct, str):
            ct = str.join('', ct)
        return ckms.utils.b64decode(ct)

    async def decrypt(self, key: KeySpecification) -> bytes:
        params: dict[str, bytes] = {}
        if self.aad: params['aad'] = ckms.utils.b64decode(self.aad)
        if self.iv: params['iv'] = ckms.utils.b64decode(self.iv)
        if self.tag: params['tag'] = ckms.utils.b64decode(self.tag)
        ct = CipherText(buf=self.get_ciphertext(), **params)
        result = key.decrypt(ct)
        if inspect.isawaitable(result):
            result = await result
        assert isinstance(result, bytes)
        return result


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Maintains credentials for external services."""
from .basecredentialrepository import BaseCredentialRepository
from .globalcredential import GlobalCredential
from .params import CredentialRepository
from .subjectcredential import SubjectCredential
from .types import ICredentialRepository


__all__: list[str] = [
    'BaseCredentialRepository',
    'CredentialRepository',
    'ICredentialRepository',
    'GlobalCredential',
    'SubjectCredential',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import functools
import logging
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import ParamSpec
from typing import TypeVar

import fastapi
import sqlalchemy.exc
import sqlalchemy.ext.asyncio

from .connectionregistry import connections


_P = ParamSpec('_P')
_T = TypeVar("_T", bound=Any)
logger: logging.Logger = logging.getLogger('cbra')


class Session(sqlalchemy.ext.asyncio.AsyncSession):
    __module__: str = 'cbra.ext.sql'

    @classmethod
    def inject(cls, name: str = 'default'):
        async def f():
            session = connections.session(name, cls=cls)
            try:
                yield session
            finally:
                await session.close()

        return fastapi.Depends(f)
    
    @staticmethod
    def retry(
        max_attempts: int,
        delay: float = 0.0,
        catch: tuple[type[BaseException], ...] = (
            sqlalchemy.exc.TimeoutError,
        ),
    ) -> Callable[
        [Callable[_P, Awaitable[_T]]],
        Callable[_P, Awaitable[_T]]
    ]:

        def decorator_factory(func: Callable[_P, Awaitable[_T]]) -> Callable[_P, Awaitable[_T]]:
            @functools.wraps(func)
            async def f(*args: Any, **kwargs: Any) -> _T:
                _attempts = kwargs.pop('_attempts', 1)
                try:
                    return await func(*args, **kwargs) # type: ignore
                except catch as e:
                    if _attempts > max_attempts:
                        raise
                    logger.warning("Recovering from %s", type(e).__name__)
                    if delay:
                        await asyncio.sleep(delay)
                    return await f(_attempts=_attempts + 1, *args, **kwargs)
            return f # type: ignore

        return decorator_factory


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from .ipolicy import IPolicy


class IPolicyManager(Protocol):
    __module__: str = 'cbra.ext.rbac.types'

    async def attach(self, __key: Any, __policy: Any) -> None:
        """Attaches `policy` to the resource specified by `key`."""
        ...

    async def get(self, __model: Any, __key: Any) -> IPolicy:
        ...